# import os
import numpy as np
# import numba
from numba import njit,types,from_dtype,typed
# from numba.typed import Dict,List
from numba.experimental import jitclass
#from numba_kdtree import KDTree

# Fit a PSF model to multiple stars in an image

PI = 3.141592653589793

from numba.pycc import CC
cc = CC('_utils_numba_static')

@cc.export('sum', 'f8(f8[:],b1)')
@cc.export('sumi', 'i8(i8[:],b1)')
def sum(data,ignore_nan=False):
    """Get the sum."""
    if ignore_nan:
        sm = nansum(data)
    else:
        sm = np.sum(data)
    return sm

@njit
@cc.export('nansum', 'f8(f8[:])')
#@cc.export('nansumi', 'i8(i8[:])')
def nansum(data):
    """ Get the sum ignoring nans """
    data = data
    gd, = np.where(np.isfinite(data)==True)
    if len(gd)==0:
        sm = np.nan
    else:
        sm = np.sum(data[gd])
    return sm

@njit
@cc.export('sum2d', 'f8[:](f8[:,:],i4,b1)')
@cc.export('sum2di', 'f8[:](i8[:,:],i4,b1)')
def sum2d(data,axis=-1,ignore_nan=False):
    """Get the sum."""
    if axis==-1:
        raise Exception('Use sum() if not using axis parameter')
    if data.ndim != 2:
        raise Exception('data must be 2D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    sm = np.zeros(loopshape[0],float)
    for i in range(loopshape[0]):
        if axis==0:
            if ignore_nan:
                sm[i] = nansum(data[:,i])
            else:
                sm[i] = np.sum(data[:,i])
        else:
            if ignore_nan:
                sm[i] = nansum(data[i,:])
            else:
                sm[i] = np.sum(data[i,:])
    return sm

@njit
@cc.export('sum3d', 'f8[:,:](f8[:,:,:],i4,b1)')
@cc.export('sum3di', 'f8[:,:](i8[:,:,:],i4,b1)')
def sum3d(data,axis=-1,ignore_nan=False):
    """Get the sum."""
    if axis==-1:
        raise Exception('Use sum() if not using axis parameter')
    if data.ndim != 3:
        raise Exception('data must be 3D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    sm = np.zeros((loopshape[0],loopshape[1]),float)
    for i in range(loopshape[0]):
        for j in range(loopshape[1]):
            if axis==0:
                if ignore_nan:
                    sm[i,j] = nansum(data[:,i,j])
                else:
                    sm[i,j] = np.sum(data[:,i,j])
            elif axis==1:
                if ignore_nan:
                    sm[i,j] = nansum(data[i,:,j])
                else:
                    sm[i,j] = np.sum(data[i,:,j])
            else:
                if ignore_nan:
                    sm[i,j] = nansum(data[i,j,:])
                else:
                    sm[i,j] = np.sum(data[i,j,:])
    return sm

@njit
@cc.export('nanmean', 'f8(f8[:])')
@cc.export('nanmeani', 'i8(i8[:])')
def nanmean(data):
    """ Get the mean ignoring nans """
    gd, = np.where(np.isfinite(data)==True)
    if len(gd)==0:
        mn = np.nan
    else:
        mn = np.mean(data[gd])
    return mn

@njit
@cc.export('mean', 'f8(f8[:],b1)')
@cc.export('meani', 'i8(i8[:],b1)')
def mean(data,ignore_nan=False):
    """Get the mean."""
    if ignore_nan:
        mn = nanmean(data)
    else:
        mn = np.mean(data)
    return mn

@njit
@cc.export('mean2d', 'f8[:](f8[:,:],i4,b1)')
@cc.export('mean2di', 'f8[:](i8[:,:],i4,b1)')
def mean2d(data,axis=-1,ignore_nan=False):
    """Get the mean."""
    if axis==-1:
        raise Exception('Use mean() if not using axis parameter')
    if data.ndim != 2:
        raise Exception('data must be 2D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    mn = np.zeros(loopshape[0],float)
    for i in range(loopshape[0]):
        if axis==0:
            if ignore_nan:
                mn[i] = nanmean(data[:,i])
            else:
                mn[i] = np.mean(data[:,i])
        else:
            if ignore_nan:
                mn[i] = nanmean(data[i,:])
            else:
                mn[i] = np.mean(data[i,:])
    return mn

@njit
@cc.export('mean3d', 'f8[:,:](f8[:,:,:],i4,b1)')
@cc.export('mean3di', 'f8[:,:](i8[:,:,:],i4,b1)')
def mean3d(data,axis=-1,ignore_nan=False):
    """Get the mean."""
    if axis==-1:
        raise Exception('Use mean() if not using axis parameter')
    if data.ndim != 3:
        raise Exception('data must be 3D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    mn = np.zeros((loopshape[0],loopshape[1]),float)
    for i in range(loopshape[0]):
        for j in range(loopshape[1]):
            if axis==0:
                if ignore_nan:
                    mn[i,j] = nanmean(data[:,i,j])
                else:
                    mn[i,j] = np.mean(data[:,i,j])
            elif axis==1:
                if ignore_nan:
                    mn[i,j] = nanmean(data[i,:,j])
                else:
                    mn[i,j] = np.mean(data[i,:,j])
            else:
                if ignore_nan:
                    mn[i,j] = nanmean(data[i,j,:])
                else:
                    mn[i,j] = np.mean(data[i,j,:])
    return mn

@njit
@cc.export('nanmedian', 'f8(f8[:])')
@cc.export('nanmediani', 'i8(i8[:])')
def nanmedian(data):
    """ Get the median ignoring nans """
    gd, = np.where(np.isfinite(data)==True)
    if len(gd)==0:
        med = np.nan
    else:
        med = np.median(data[gd])
    return med

@njit
@cc.export('median', 'f8(f8[:],b1)')
@cc.export('mediani', 'i8(i8[:],b1)')
def median(data,ignore_nan=False):
    """Get the median."""
    if ignore_nan:
        med = nanmedian(data)
    else:
        med = np.median(data)
    return med

@njit
@cc.export('median2d', 'f8[:](f8[:,:],i4,b1)')
@cc.export('median2di', 'f8[:](i8[:,:],i4,b1)')
def median2d(data,axis=-1,ignore_nan=False):
    """Get the median."""
    if axis==-1:
        raise Exception('Use median() if not using axis parameter')
    if data.ndim != 2:
        raise Exception('data must be 2D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    med = np.zeros(loopshape[0],float)
    for i in range(loopshape[0]):
        if axis==0:
            if ignore_nan:
                med[i] = nanmedian(data[:,i])
            else:
                med[i] = np.median(data[:,i])
        else:
            if ignore_nan:
                med[i] = nanmedian(data[i,:])
            else:
                med[i] = np.median(data[i,:])
    return med

@njit
@cc.export('median3d', 'f8[:,:](f8[:,:,:],i4,b1)')
@cc.export('median3di', 'f8[:,:](i8[:,:,:],i4,b1)')
def median3d(data,axis=-1,ignore_nan=False):
    """Get the median."""
    if axis==-1:
        raise Exception('Use median() if not using axis parameter')
    if data.ndim != 3:
        raise Exception('data must be 3D')
    ndim = data.ndim
    shape = data.shape
    loopdims = np.delete(np.arange(ndim),axis)
    loopshape = np.delete(np.array(shape),axis)
    med = np.zeros((loopshape[0],loopshape[1]),float)
    for i in range(loopshape[0]):
        for j in range(loopshape[1]):
            if axis==0:
                if ignore_nan:
                    med[i,j] = nanmedian(data[:,i,j])
                else:
                    med[i,j] = np.median(data[:,i,j])
            elif axis==1:
                if ignore_nan:
                    med[i,j] = nanmedian(data[i,:,j])
                else:
                    med[i,j] = np.median(data[i,:,j])
            else:
                if ignore_nan:
                    med[i,j] = nanmedian(data[i,j,:])
                else:
                    med[i,j] = np.median(data[i,j,:])
    return med

@njit
@cc.export('mad', 'f8(f8[:],b1,b1)')
@cc.export('madi', 'i8(i8[:],b1,b1)')
def mad(data, ignore_nan=True, zero=False):
    """ Calculate the median absolute deviation of an array."""
    # With median reference point
    if zero==False:
        if ignore_nan:
            ref = np.nanmedian(data)
            result= np.median(np.abs(data-ref))
        else:
            ref = np.median(data)
            result = nanmedian(np.abs(data-ref))
    # Using zero as reference point
    else:
        if ignore_nan:
            result= np.median(np.abs(data))
        else:
            ref = np.median(data)
            result = nanmedian(np.abs(data))
    return result * 1.482602218505602

@njit
@cc.export('mad2d', 'f8[:](f8[:,:],i4,b1,b1)')
@cc.export('mad2di', 'f8[:](i8[:,:],i4,b1,b1)')
def mad2d(data, axis=-1, ignore_nan=True, zero=False):
    """ Calculate the median absolute deviation of an array."""
    if axis==-1:
        raise Exception('Use mad() if not using axis parameter')
    if data.ndim != 2:
        raise Exception('data must be 2D')
    # With median reference point
    if zero==False:
        if ignore_nan:
            ref = median2d(data,axis=axis,ignore_nan=True)
            newshape = np.array(data.shape)
            newshape[axis] = -1
            newshape = (newshape[0],newshape[1])
            resid = data-ref.reshape(newshape)
            result = median2d(np.abs(resid),axis=axis,ignore_nan=True)
        else:
            ref = median2d(data,axis=axis)
            newshape = np.array(data.shape)
            newshape[axis] = -1
            newshape = (newshape[0],newshape[1])
            resid = data-ref.reshape(newshape)
            result = median2d(np.abs(resid),axis=axis)
    # Using zero as reference point
    else:
        if ignore_nan:
            result = median2d(np.abs(data),axis=axis,ignore_nan=True)
        else:
            result = median2d(np.abs(data),axis=axis)

    return result * 1.482602218505602

@njit
@cc.export('mad3d', 'f8[:,:](f8[:,:,:],i4,b1,b1)')
@cc.export('mad3di', 'f8[:,:](i8[:,:,:],i4,b1,b1)')
def mad3d(data, axis=-1, ignore_nan=True, zero=False):
    """ Calculate the median absolute deviation of an array."""
    if axis==-1:
        raise Exception('Use mad() if not using axis parameter')
    if data.ndim != 3:
        raise Exception('data must be 3D')
    # With median reference point
    if zero==False:
        if ignore_nan:
            ref = median3d(data,axis=axis,ignore_nan=True)
            newshape = np.array(data.shape)
            newshape[axis] = -1
            newshape = (newshape[0],newshape[1],newshape[2])
            resid = data-ref.reshape(newshape)
            result = median3d(np.abs(resid),axis=axis,ignore_nan=True)
        else:
            ref = median3d(data,axis=axis)
            newshape = np.array(data.shape)
            newshape[axis] = -1
            newshape = (newshape[0],newshape[1],newshape[2])
            resid = data-ref.reshape(newshape)
            result = median3d(np.abs(resid),axis=axis)
    # Using zero as reference point
    else:
        if ignore_nan:
            result = median3d(np.abs(data),axis=axis,ignore_nan=True)
        else:
            result = median3d(np.abs(data),axis=axis)
    return result * 1.482602218505602

@njit
@cc.export('quadratic_bisector', 'f8(f8[:],f8[:])')
@cc.export('quadratic_bisectori', 'f8(i8[:],i8[:])')
def quadratic_bisector(x,y):
    """ Calculate the axis of symmetric or bisector of parabola"""
    #https://www.azdhs.gov/documents/preparedness/state-laboratory/lab-licensure-certification/technical-resources/
    #    calibration-training/12-quadratic-least-squares-regression-calib.pdf
    #quadratic regression statistical equation
    n = len(x)
    if n<3:
        return np.nan
    Sxx = np.sum(x**2) - np.sum(x)**2/n
    Sxy = np.sum(x*y) - np.sum(x)*np.sum(y)/n
    Sxx2 = np.sum(x**3) - np.sum(x)*np.sum(x**2)/n
    Sx2y = np.sum(x**2 * y) - np.sum(x**2)*np.sum(y)/n
    Sx2x2 = np.sum(x**4) - np.sum(x**2)**2/n
    #a = ( S(x^2*y)*S(xx)-S(xy)*S(xx^2) ) / ( S(xx)*S(x^2x^2) - S(xx^2)^2 )
    #b = ( S(xy)*S(x^2x^2) - S(x^2y)*S(xx^2) ) / ( S(xx)*S(x^2x^2) - S(xx^2)^2 )
    denom = Sxx*Sx2x2 - Sxx2**2
    if denom==0:
        return np.nan
    a = ( Sx2y*Sxx - Sxy*Sxx2 ) / denom
    b = ( Sxy*Sx2x2 - Sx2y*Sxx2 ) / denom
    if a==0:
        return np.nan
    return -b/(2*a)

@njit
@cc.export('meshgrid', '(f8[:],f8[:])')
@cc.export('meshgridi', '(i8[:],i8[:])')
def meshgrid(x,y):
    """ Implementation of numpy's meshgrid function."""
    nx = len(x)
    ny = len(y)
    dtype = np.array(x[0]*y[0]).dtype
    xx = np.zeros((ny,nx),dtype)
    for i in range(ny):
        xx[i,:] = x
    yy = np.zeros((ny,nx),dtype)
    for i in range(nx):
        yy[:,i] = y
    return xx,yy


# @njit
# @cc.export('meshgridi', '(i8[:],i8[:])')
# def meshgridi(x,y):
#     """ Implementation of numpy's meshgrid function."""
#     nx = len(x)
#     ny = len(y)
#     dtype = np.array(x[0]*y[0]).dtype
#     xx = np.zeros((ny,nx),dtype)
#     for i in range(ny):
#         xx[i,:] = x
#     yy = np.zeros((ny,nx),dtype)
#     for i in range(nx):
#         yy[:,i] = y
#     return xx,yy

@njit
@cc.export('aclip', 'f8[:](f8[:],f8,f8)')
def aclip(val,minval,maxval):
    newvals = np.zeros(len(val),float)
    for i in range(len(val)):
        if val[i] < minval:
            nval = minval
        elif val[i] > maxval:
            nval = maxval
        else:
            nval = val[i]
        newvals[i] = nval
    return newvals

@njit
@cc.export('clip', 'f8(f8,f8,f8)')
@cc.export('clipi', 'i8(i8,i8,i8)')
def clip(val,minval,maxval):
    if val < minval:
        nval = minval
    elif val > maxval:
        nval = maxval
    else:
        nval = val
    return nval

@njit
@cc.export('drop_imag', 'f8(f8)')
def drop_imag(z):
    EPSILON = 1e-07    
    if np.abs(z.imag) <= EPSILON:
        z = z.real
    return z

@njit('f8(f8)')
@cc.export('gamma', 'f8(f8)')
def gamma(z):
    # Gamma function for a single z value
    # Using the Lanczos approximation
    # https://en.wikipedia.org/wiki/Lanczos_approximation
    g = 7
    n = 9
    p = [
        0.99999999999980993,
        676.5203681218851,
        -1259.1392167224028,
        771.32342877765313,
        -176.61502916214059,
        12.507343278686905,
        -0.13857109526572012,
        9.9843695780195716e-6,
        1.5056327351493116e-7
    ]
    if z < 0.5:
        y = PI / (np.sin(PI * z) * gamma(1 - z))  # Reflection formula
    else:
        z -= 1
        x = p[0]
        for i in range(1, len(p)):
            x += p[i] / (z + i)
        t = z + g + 0.5
        y = np.sqrt(2 * PI) * t ** (z + 0.5) * np.exp(-t) * x
    return y

@njit
@cc.export('gammaincinv05', 'f8(f8)')
def gammaincinv05(a):
    """ gammaincinv(a,0.5) """
    n = np.array([1.00e-03, 1.12e-01, 2.23e-01, 3.34e-01, 4.45e-01, 5.56e-01,
                  6.67e-01, 7.78e-01, 8.89e-01, 1.00e+00, 2.00e+00, 3.00e+00,
                  4.00e+00, 5.00e+00, 6.00e+00, 7.00e+00, 8.00e+00, 9.00e+00,
                  1.00e+01, 1.10e+01, 1.20e+01])

    y = np.array([5.24420641e-302, 1.25897478e-003, 3.03558724e-002, 9.59815712e-002,
                  1.81209305e-001, 2.76343765e-001, 3.76863377e-001, 4.80541354e-001,
                  5.86193928e-001, 6.93147181e-001, 1.67834699e+000, 2.67406031e+000,
                  3.67206075e+000, 4.67090888e+000, 5.67016119e+000, 6.66963707e+000,
                  7.66924944e+000, 8.66895118e+000, 9.66871461e+000, 1.06685224e+001,
                  1.16683632e+001])
    # Lower edge
    if a < n[0]:
        out = 0.0
    # Upper edge
    elif a > n[-1]:
        # linear approximation to large values
        # coef = np.array([ 0.99984075, -0.32972584])
        out = 0.99984075*a-0.32972584
    # Interpolate values
    else:
        ind = np.searchsorted(n,a)
        # exact match
        if n[ind]==a:
            out = y[ind]
        # At beginning
        elif ind==0:
            slp = (y[1]-y[0])/(n[1]-n[0])
            out = slp*(a-n[0])+y[0]
        else:
            slp = (y[ind]-y[ind-1])/(n[ind]-n[ind-1])
            out = slp*(a-n[ind-1])+y[ind-1]
            
    return out

    
@njit
@cc.export('linearinterp', 'f8(f8[:,:],f8,f8)')
@cc.export('linearinterpi', 'f8(i8[:,:],i8,i8)')
def linearinterp(data,x,y):
    """
    Linear interpolation.

    Parameters
    ----------
    data : numpy array
      The data to use for interpolation.
    x : float
      The X-value at which to perform the interpolation.
    y : float
      The Y-value at which to perform the interpolation.

    Returns
    -------
    f : float
       The interpolated value.
    
    Examples
    --------

    f = linearinterp(data,x,y)

    """

    
    ny,nx = data.shape

    # Out of bounds
    if x<0 or x>(nx-1) or y<0 or y>(ny-1):
        return 0.0

    x1 = int(x)
    if x1==nx-1:
        x1 -= 1
    x2 = x1+1
    y1 = int(y)
    if y1==ny-1:
        y1 -= 1
    y2 = y1+1

    f11 = data[y1,x1]
    f12 = data[y2,x1]
    f21 = data[y1,x2]
    f22 = data[y2,x2]
    
    # weighted mean
    # denom = (x2-x1)*(y2-y1) = 1
    w11 = (x2-x)*(y2-y)
    w12 = (x2-x)*(y-y1)
    w21 = (x-x1)*(y2-y)
    w22 = (x-x1)*(y-y1)
    f = w11*f11+w12*f12+w21*f21+w22*f22

    return f

@njit
@cc.export('alinearinterp', '(f8[:,:],f8[:],f8[:])')
@cc.export('alinearinterpi', '(f8[:,:],i8[:],i8[:])')
def alinearinterp(data,x,y):
    """
    Linear interpolation.

    Parameters
    ----------
    data : numpy array
      The data to use for interpolation.
    x : numpy array
      Array of X-value at which to perform the interpolation.
    y : numpy array
      Array of Y-value at which to perform the interpolation.

    Returns
    -------
    f : numpy array
       The interpolated values.
    
    Examples
    --------

    f = alinearinterp(data,x,y)

    """
    npix = len(x)
    f = np.zeros(npix,float)
    for i in range(npix):
        f[i] = linearinterp(data,x[i],y[i])
    return f

@njit
@cc.export('inverse', 'f8[:,:](f8[:,:])')
def inverse(a):
    """ Safely take the inverse of a square 2D matrix."""
    # This checks for zeros on the diagonal and "fixes" them.
    
    # If one of the dimensions is zero in the R matrix [Npars,Npars]
    # then replace it with a "dummy" value.  A large value in R
    # will give a small value in inverse of R.
    #badpar, = np.where(np.abs(np.diag(a))<sys.float_info.min)
    badpar = (np.abs(np.diag(a))<2e-300)
    if np.sum(badpar)>0:
        a[badpar] = 1e10
    ainv = np.linalg.inv(a)
    # What if the inverse fails???
    # can we catch it in numba
    # Fix values
    a[badpar] = 0  # put values back
    ainv[badpar] = 0
    
    return ainv

@njit
@cc.export('qr_jac_solve', '(f8[:,:],f8[:],f8[:])')
def qr_jac_solve(jac,resid,weight=None):
    """ Solve part of a non-linear least squares equation using QR decomposition
        using the Jacobian."""
    # jac: Jacobian matrix, first derivatives, [Npix, Npars]
    # resid: residuals [Npix]
    # weight: weights, ~1/error**2 [Npix]

    #if weight is not None and weight.flags['C_CONTIGUOUS']==False:
    #    # Make array contiguous
    if weight is not None:
        weight = np.ascontiguousarray(weight)
    #jac = np.ascontiguousarray(jac)
    #resid = np.ascontiguousarray(resid)
        
    # QR decomposition
    if weight is None:
        q,r = np.linalg.qr(jac)
        rinv = inverse(r)
        dbeta = rinv @ (q.T @ resid)
    # Weights input, multiply resid and jac by weights        
    else:
        weight = np.ascontiguousarray(weight)
        q,r = np.linalg.qr( jac * weight.reshape(-1,1) )
        rinv = inverse(r)
        dbeta = rinv @ (q.T @ (resid*weight))
        
    return dbeta

@njit
@cc.export('jac_covariance', '(f8[:,:],f8[:],f8[:])')
def jac_covariance(jac,resid,wt):
    """ Determine the covariance matrix. """
    
    npix,npars = jac.shape

    # Make array contiguous
    wt = np.ascontiguousarray(wt)
    
    # Weights
    #   If weighted least-squares then
    #   J.T * W * J
    #   where W = I/sig_i**2
    if wt is not None:
        wt2 = wt.reshape(-1,1) + np.zeros(npars)
        hess = jac.T @ (wt2 * jac)
    else:
        hess = jac.T @ jac  # not weighted

    # cov = H-1, covariance matrix is inverse of Hessian matrix
    cov_orig = inverse(hess)
    
    # Rescale to get an unbiased estimate
    # cov_scaled = cov * (RSS/(m-n)), where m=number of measurements, n=number of parameters
    # RSS = residual sum of squares
    #  using rss gives values consistent with what curve_fit returns
    # Use chi-squared, since we are doing the weighted least-squares and weighted Hessian
    if wt is not None:
        chisq = np.sum(resid**2 * wt)
    else:
        chisq = np.sum(resid**2)
    dof = npix-npars
    if dof<=0:
        dof = 1
    cov = cov_orig * (chisq/dof)  # what MPFITFUN suggests, but very small
        
    return cov

# @njit
# @cc.export('checkbounds', '(f8[:],f8[:,:])')
# def checkbounds(pars,bounds):
#     """ Check the parameters against the bounds."""
#     # 0 means it's fine
#     # 1 means it's beyond the lower bound
#     # 2 means it's beyond the upper bound
#     npars = len(pars)
#     lbounds,ubounds = bounds[:,0],bounds[:,1]
#     check = np.zeros(npars,np.int32)
#     badlow, = np.where(pars<=lbounds)
#     if len(badlow)>0:
#         check[badlow] = 1
#     badhigh, = np.where(pars>=ubounds)
#     if len(badhigh)>0:
#         check[badhigh] = 2
#     return check

@njit
@cc.export('checkbounds', '(f8[:],UniTuple(f8[:],2))')
def checkbounds(pars,bounds):
    """ Check the parameters against the bounds."""
    # 0 means it's fine
    # 1 means it's beyond the lower bound
    # 2 means it's beyond the upper bound
    npars = len(pars)
    lbounds,ubounds = bounds
    check = np.zeros(npars,np.int32)
    badlo, = np.where(pars<=lbounds)
    if len(badlo)>0:
        check[badlo] = 1
    badhi, = np.where(pars>=ubounds)
    if len(badhi):
        check[badhi] = 2
    return check

@njit
@cc.export('limbounds', '(f8[:],UniTuple(f8[:],2))')
def limbounds(pars,bounds):
    """ Limit the parameters to the boundaries."""
    lbounds,ubounds = bounds
    outpars = np.minimum(np.maximum(pars,lbounds),ubounds)
    return outpars

@njit
@cc.export('limsteps', '(f8[:],f8[:])')
def limsteps(steps,maxsteps):
    """ Limit the parameter steps to maximum step sizes."""
    signs = np.sign(steps)
    outsteps = np.minimum(np.abs(steps),maxsteps)
    outsteps *= signs
    return outsteps

# @njit
# @cc.export('newpars', '(f8[:],f8[:],f8[:,:],f8[:])')
# def newpars(pars,steps,bounds=None,maxsteps=None):
#     """ Return new parameters that fit the constraints."""
#     # Limit the steps to maxsteps
#     if maxsteps is not None:
#         limited_steps = limsteps(steps,maxsteps)
#     else:
#         limited_steps = steps

#     # No bounds input
#     if bounds is None:
#         return pars+limited_steps
        
#     # Make sure that these don't cross the boundaries
#     lbounds,ubounds = bounds[:,0],bounds[:,1]
#     check = checkbounds(pars+limited_steps,bounds)
#     # Reduce step size for any parameters to go beyond the boundaries
#     badpars = (check!=0)
#     # reduce the step sizes until they are within bounds
#     newsteps = limited_steps.copy()
#     count = 0
#     maxiter = 2
#     while (np.sum(badpars)>0 and count<=maxiter):
#         newsteps[badpars] /= 2
#         newcheck = checkbounds(pars+newsteps,bounds)
#         badpars = (newcheck!=0)
#         count += 1
            
#     # Final parameters
#     newpars = pars + newsteps
                
#     # Make sure to limit them to the boundaries
#     check = checkbounds(newpars,bounds)
#     badpars = (check!=0)
#     if np.sum(badpars)>0:
#         # add a tiny offset so it doesn't fit right on the boundary
#         newpars = np.minimum(np.maximum(newpars,lbounds+1e-30),ubounds-1e-30)
#     return newpars

@njit
@cc.export('poly2d', '(f8[:,:],f8[:])')
def poly2d(xdata,pars):
    """ model of 2D linear polynomial."""
    x = xdata[:,0]
    y = xdata[:,1]
    return pars[0]+pars[1]*x+pars[2]*y+pars[3]*x*y

@njit
@cc.export('jacpoly2d', '(f8[:,:],f8[:])')
def jacpoly2d(xdata,pars):
    """ jacobian of 2D linear polynomial."""
    x = xdata[:,0]
    y = xdata[:,1]
    nx = len(x)
    # Model
    m = pars[0]+pars[1]*x+pars[2]*y+pars[3]*x*y
    # Jacobian, partical derivatives wrt the parameters
    jac = np.zeros((nx,4),float)
    jac[:,0] = 1    # constant coefficient
    jac[:,1] = x    # x-coefficient
    jac[:,2] = y    # y-coefficient
    jac[:,3] = x*y  # xy-coefficient
    return m,jac

@njit
@cc.export('poly2dfit', '(f8[:],f8[:],f8[:],f8[:],i4,f8,b1)')
def poly2dfit(x,y,data,error,maxiter=2,minpercdiff=0.5,verbose=False):
    """ Fit a 2D linear function to data robustly."""
    ndata = len(data)
    if ndata<4:
        raise Exception('Need at least 4 data points for poly2dfit')
    gd1, = np.where(np.isfinite(data))
    if len(gd1)<4:
        raise Exception('Need at least 4 good data points for poly2dfit')
    xdata = np.zeros((len(gd1),2),float)
    xdata[:,0] = x[gd1]
    xdata[:,1] = y[gd1]
    initpar = np.zeros(4,float)
    med = np.median(data[gd1])
    sig = mad(data[gd1])
    gd2, = np.where( (np.abs(data-med)<3*sig) & np.isfinite(data))
    if len(gd1)>=4 and len(gd2)<4:
        gd = gd1
    else:
        gd = gd2
    initpar[0] = med
    xdata = np.zeros((len(gd),2),float)
    xdata[:,0] = x[gd]
    xdata[:,1] = y[gd]
    data1 = data[gd]
    error1 = error[gd]

    # Do the fit
    # Iterate
    count = 0
    bestpar = initpar.copy()
    maxpercdiff = 1e10
    # maxsteps = None
    #wt = 1.0/error1.ravel()**2
    wt = np.ones(len(error1),float)  # unweighted
    while (count<maxiter and maxpercdiff>minpercdiff):
        # Use Cholesky, QR or SVD to solve linear system of equations
        m,j = jacpoly2d(xdata,bestpar)
        dy = data1.ravel()-m.ravel()
        # Solve Jacobian
        #if error is not None:
        #dbeta = qr_jac_solve(j,dy,weight=wt)
        dbeta = qr_jac_solve(j,dy,wt)
        #else:
        #    dbeta = qr_jac_solve(j,dy)
        dbeta[~np.isfinite(dbeta)] = 0.0  # deal with NaNs

        # -add "shift cutting" and "line search" in the least squares method
        # basically scale the beta vector to find the best match.
        # check this out
        # https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.line_search.html
        
        # Update parameters
        oldpar = bestpar.copy()
        # limit the steps to the maximum step sizes and boundaries
        #if bounds is not None or maxsteps is not None:
        #    bestpar = newpars(bestpar,dbeta,bounds,maxsteps)
        #else:
        bestpar += dbeta
        # Check differences and changes
        diff = np.abs(bestpar-oldpar)
        denom = np.maximum(np.abs(oldpar.copy()),0.0001)
        percdiff = diff.copy()/denom*100  # percent differences
        maxpercdiff = np.max(percdiff)
                
        if verbose:
            print('N = ',count)
            print('bestpars = ',bestpar)
            print('dbeta = ',dbeta)
                
        count += 1

    # Get covariance and errors
    m,j = jacpoly2d(xdata,bestpar)
    dy = data1.ravel()-m.ravel()
    cov = jac_covariance(j,dy,wt)
    perror = np.sqrt(np.diag(cov))

    return bestpar,perror,cov


spec = [
    ('_inputtype', types.int64),
    ('_data', types.float64[:]),
    ('ndata', types.int32),
    ('_sindex', types.int64[:]),
    ('_sdata', types.float64[:]),
    ('_num', types.int64[:]),
    ('_values', types.float64[:]),
    ('_lo', types.int64[:]),
    ('_hi', types.int64[:]),
    ('_invindex', types.int64[:]),
]
@jitclass(spec)
class Index(object):

    def __init__(self, arr):
        if len(arr)==0:
            raise ValueError('arr has no elements')
        aint32 = np.zeros(10,np.int32)
        aint64 = np.zeros(10,np.int64)
        if arr.dtype is aint32.dtype or arr.dtype is aint64.dtype:
            self._inputtype = 2     # int
        else:
            self._inputtype = 1     # float
        data = np.zeros(len(arr),float)   # make sure they are float
        data[:] = arr
        self._data = data
        self.ndata = len(data)
        self._sindex = np.argsort(self._data)
        self._sdata = data.copy()[self._sindex]
        brklo, = np.where(self._sdata != np.roll(self._sdata,1))
        nbrk = len(brklo)
        if nbrk>0:
            brkhi = np.hstack((brklo[1:nbrk]-1,np.array([self.ndata-1])))
            num = brkhi-brklo+1
            self._num = num
            self._values = self._sdata[brklo]
            self._lo = brklo
            self._hi = brkhi
        else:
            self._num = np.array([1])
            self._values = np.array([data[0]])
            self._lo = np.array([0])
            self._hi = np.array([self.ndata-1])
        self._invindex = np.zeros(len(arr),np.int64)-1
            
    @property
    def data(self):
        return self._data

    @property
    def index(self):
        # unique index
        return self._sindex[self._lo]
    
    @property
    def num(self):
        return self._num

    def __len__(self):
        return len(self._num)
    
    def __getitem__(self,item):
        return self._values[item]
    
    @property
    def values(self):
        return self._values
    
    def get(self,item):
        lo = self._lo[item]
        hi = self._hi[item]
        n = hi-lo+1
        return self._sdata[lo:hi+1]

    def getindex(self,item):
        lo = self._lo[item]
        hi = self._hi[item]
        n = hi-lo+1
        return self._sindex[lo:hi+1]

    @property
    def invindex(self):
        # Construct inverse index
        #  takes you from the original pixels to the unique ones (index.values)
        if self._invindex[0]==-1:
            # Loop over the unique values
            for i in range(len(self)):
                ind = self.getindex(i)   # index array for each unique value
                self._invindex[ind] = i
        return self._invindex

@njit
@cc.export('ravel_multi_index', '(UniTuple(i8[:],2),UniTuple(i8,2))')
@cc.export('ravel_multi_index3', '(UniTuple(i8[:],3),UniTuple(i8,3))')
def ravel_multi_index(multi_index,imshape):
    """ ravel indices"""
    # multi_index: tuple of integer arrays

    n = len(multi_index[0])
    ndim = len(imshape)
    factor = np.zeros(ndim,np.int64)
    factor[-1] = 1
    for i in np.arange(ndim-2,-1,-1):
        factor[i] = factor[i+1]*imshape[i+1]
    
    # Loop over points
    index = np.zeros(n,np.int64)
    for i in range(n):
        for j in range(ndim):
            index[i] += multi_index[j][i]*factor[j]
    
    return index

@njit
@cc.export('unravel_index', '(i8[:],UniTuple(i8,2))')
def unravel_index(indices,imshape):
    """ return multi-dimensional index. """

    n = len(indices)
    ndim = len(imshape)
    factor = np.zeros(ndim,np.int64)
    factor[-1] = 1
    for i in np.arange(ndim-2,-1,-1):
        factor[i] = factor[i+1]*imshape[i+1]
    
    unraveled_coords = np.zeros((n,ndim),np.int64)
    for i in range(n):
        left = indices[i]
        for j in range(ndim):
            unraveled_coords[i,j] = left//factor[j]
            left -= unraveled_coords[i,j]*factor[j]

    return unraveled_coords

# @njit
# @cc.export('unique_indexf', '(f8[:],)')
# @cc.export('unique_indexi', '(i8[:],)')
# def unique_index(array):
#     """ return unique values, index and reverse indez."""
#     index = Index(array)
#     uvals = array[index.index]
#     # Unique index
#     uindex = index.index
#     # Inverse index
#     #   the inverse index list takes you from the original/duplicated pixels
#     #   to the unique ones
#     invindex = index.invindex
#     return uvals,uindex,invindex

@njit
@cc.export('unique_index', '(f8[:],)')
def unique_indexf(array):
    data = np.zeros(len(array),float)   # make sure they are float
    data[:] = array
    ndata = len(data)
    sindex = np.argsort(data)
    sdata = data.copy()[sindex]
    brklo, = np.where(sdata != np.roll(sdata,1))
    nbrk = len(brklo)
    if nbrk>0:
        brkhi = np.hstack((brklo[1:nbrk]-1,np.array([ndata-1])))
        num = brkhi-brklo+1
        values = sdata[brklo]
        lo = brklo
        hi = brkhi
    else:
        num = np.array([1])
        values = np.array([data[0]])
        lo = np.array([0])
        hi = np.array([ndata-1])
    # Construct inverse index
    #  takes you from the original pixels to the unique ones (index.values)
    invindex = np.zeros(len(array),np.int64)-1
    # Loop over the unique values
    for i in range(len(num)):
        #ind = self.getindex(i)   # index array for each unique value
        lo1 = lo[i]
        hi1 = hi[i]
        n = hi1-lo1+1
        ind = sindex[lo1:hi1+1]
        invindex[ind] = i
    # Unique index
    uindex = sindex[lo]
    # Unique values
    uvals = data[uindex]
    return uvals,uindex,invindex


@njit
@cc.export('unique_indexi', '(i8[:],)')
def unique_index(array):
    data = np.zeros(len(array),np.int64)   # make sure they are float
    data[:] = array
    ndata = len(data)
    sindex = np.argsort(data)
    sdata = data.copy()[sindex]
    brklo, = np.where(sdata != np.roll(sdata,1))
    nbrk = len(brklo)
    if nbrk>0:
        brkhi = np.hstack((brklo[1:nbrk]-1,np.array([ndata-1])))
        num = brkhi-brklo+1
        values = sdata[brklo]
        lo = brklo
        hi = brkhi
    else:
        num = np.array([1])
        values = np.array([data[0]])
        lo = np.array([0])
        hi = np.array([ndata-1])
    # Construct inverse index
    #  takes you from the original pixels to the unique ones (index.values)
    invindex = np.zeros(len(array),np.int64)-1
    # Loop over the unique values
    for i in range(len(num)):
        #ind = self.getindex(i)   # index array for each unique value
        lo1 = lo[i]
        hi1 = hi[i]
        n = hi1-lo1+1
        ind = sindex[lo1:hi1+1]
        invindex[ind] = i
    # Unique index
    uindex = sindex[lo]
    # Unique values
    uvals = data[uindex]
    return uvals,uindex,invindex


@njit
@cc.export('smooth2d', '(f8[:,:],i8,i4)')
def smooth2d(im,binsize,med=0):
    """ Do smoothing of a 2D image."""
    # mean by default
    smim = im.copy()
    ny,nx = im.shape
    ny2 = int(np.ceil(ny / binsize))
    nx2 = int(np.ceil(nx / binsize))
    for i in range(nx2):
        for j in range(ny2):
            x1 = i*binsize
            x2 = x1+binsize
            if x2 > nx: x2=nx
            y1 = j*binsize
            y2 = y1+binsize
            if y2 > ny: y2=ny
            if med==1:
                sm = np.median(im[y1:y2,x1:x2])
            else:
                sm = np.mean(im[y1:y2,x1:x2])
            smim[y1:y2,x1:x2] = sm
    return smim

@njit
@cc.export('skygrid', '(f8[:,:],i8,i4,i4)')
def skygrid(im,binsize,tot=0,med=1):
    """ Estimate the background."""
    #binsize = 200
    ny,nx = im.shape
    ny2 = ny // binsize
    nx2 = nx // binsize
    bgim = np.zeros((ny2,nx2),float)
    nsample = np.minimum(1000,binsize*binsize)
    sample = np.random.randint(0,binsize*binsize-1,nsample)
    for i in range(nx2):
        for j in range(ny2):
            x1 = i*binsize
            x2 = x1+binsize
            if x2 > nx: x2=nx
            y1 = j*binsize
            y2 = y1+binsize
            if y2 > ny: y2=ny
            if tot==1:
                bgim[j,i] = np.sum(im[y1:y2,x1:x2].ravel()[sample])
            elif med==1:
                bgim[j,i] = np.median(im[y1:y2,x1:x2].ravel()[sample])
            else:
                bgim[j,i] = np.mean(im[y1:y2,x1:x2].ravel()[sample])
    return bgim

@njit
@cc.export('skyinterp', '(f8[:,:],f8[:,:],i8)')
def skyinterp(binim,fullim,binsize):
    """ linear interpolation"""
    ny,nx = fullim.shape
    ny2 = ny // binsize
    nx2 = nx // binsize
    hbinsize = int(0.5*binsize)

    # Calculate midpoint positions
    xx = np.arange(nx2)*binsize+hbinsize
    yy = np.arange(ny2)*binsize+hbinsize
    
    for i in range(ny2-1):
        for j in range(nx2-1):
            y1 = i*binsize+hbinsize
            y2 = y1+binsize
            x1 = j*binsize+hbinsize
            x2 = x1+binsize
            f11 = binim[i,j]
            f12 = binim[i+1,j]
            f21 = binim[i,j+1]
            f22 = binim[i+1,j+1]
            denom = binsize*binsize
            for k in range(binsize):
                for l in range(binsize):
                    x = x1+k
                    y = y1+l
                    # weighted mean
                    #denom = (x2-x1)*(y2-y1)
                    w11 = (x2-x)*(y2-y)/denom
                    w12 = (x2-x)*(y-y1)/denom
                    w21 = (x-x1)*(y2-y)/denom
                    w22 = (x-x1)*(y-y1)/denom
                    f = w11*f11+w12*f12+w21*f21+w22*f22
                    fullim[y,x] = f

    # Do the edges
    for i in range(hbinsize,nx2*binsize-hbinsize):
        # bottom
        for j in range(hbinsize):
            fullim[j,i] = fullim[hbinsize,i]
        # top
        for j in range(ny2*binsize-hbinsize-1,ny):
            fullim[j,i] = fullim[ny2*binsize-hbinsize-2,i]
    for j in np.arange(hbinsize,ny2*binsize-hbinsize):
        # left
        for i in range(hbinsize):
            fullim[j,i] = fullim[j,hbinsize]
        # right
        for i in range(nx2*binsize-hbinsize-1,nx):
            fullim[j,i] = fullim[j,nx2*binsize-hbinsize-2]
    # Do the corners
    for i in range(hbinsize):
        # bottom-left
        for j in range(hbinsize):
            fullim[j,i] = fullim[hbinsize,hbinsize]
        # top-left
        for j in range(ny2*binsize-hbinsize-1,ny):
            fullim[j,i] = fullim[ny2*binsize-hbinsize-2,hbinsize]
    for i in range(nx2*binsize-hbinsize-1,nx):
        # bottom-right
        for j in range(hbinsize):
            fullim[j,i] = fullim[hbinsize,nx2*binsize-hbinsize-2]
        # top-right
        for j in range(ny2*binsize-hbinsize-1,ny):
            fullim[j,i] = fullim[ny2*binsize-hbinsize-2,nx2*binsize-hbinsize-2]

    return fullim
                    

@njit
@cc.export('sky', '(f8[:,:],i8)')
def sky(im,binsize=0):
    ny,nx = im.shape

    # Figure out best binsize
    if binsize <= 0:
        binsize = np.min(np.array([ny//20,nx//20]))
    binsize = np.maximum(binsize,20)
    ny2 = ny // binsize
    nx2 = nx // binsize

    # Bin in a grid
    bgimbin = skygrid(im,binsize,0,1)

    # Outlier rejection
    # median smoothing
    medbin = 3
    if nx2<medbin or ny2<medbin:
        medbin = np.min(np.array([nx2,ny2]))
    smbgimbin = smooth2d(bgimbin,medbin,1)  # median smoothing
    
    # Linearly interpolate
    bgim = np.zeros(im.shape,np.float64)+np.median(bgimbin)
    bgim = skyinterp(smbgimbin,bgim,binsize)

    return bgim

# @njit
# @cc.export('numba_sky2', '(f8[:],f8[:],f8[:,:],f8[:])')
# def numba_sky2(im,binsize,tot,med):
#     """ Estimate the background."""
#     ny,nx = im.shape
#     ny2 = ny // binsize
#     nx2 = nx // binsize
#     bgim = np.zeros((ny2,nx2),float)
#     #sample = np.random.randint(0,binsize*binsize-1,1000)
#     xsample = np.random.randint(0,binsize-1,1000)
#     ysample = np.random.randint(0,binsize-1,1000)
#     data = np.zeros(1000,float)
#     for i in range(ny2):
#         for j in range(nx2):
#             x1 = i*binsize
#             x2 = x1+binsize
#             if x2 > nx: x2=nx
#             y1 = j*binsize
#             y2 = y1+binsize
#             if y2 > ny: y2=ny
#             for k in range(1000):
#                 data[k] = im[y1+ysample[k],x1+xsample[k]]
#             if tot==1:
#                 bgim[j,i] = np.sum(data)
#             elif med==1:
#                 bgim[j,i] = np.median(data)
#             else:
#                 bgim[j,i] = np.mean(data)
#     #import pdb; pdb.set_trace()

#     return bgim


# @njit
# @cc.export('detection', '(f8[:,:],i4)')
# def detection(im,nsig=10):
#     """  Detect peaks """

#     # just looping over the 9K x 9K array
#     # takes 1.3 sec

#     # bin 2x2 as a crude initial smoothing
#     imbin = dln.rebin(im,binsize=(2,2))
    
#     sig = sigma(imbin)
#     xpeak,ypeak,count = detectpeaks(imbin,sig,nsig)
#     xpeak = xpeak[:count]*2
#     ypeak = ypeak[:count]*2
    
#     return xpeak,ypeak

@njit
@cc.export('detectpeaks', '(f8[:,:],f8,i8)')
def detectpeaks(im,sig,nsig):
    """ Detect peaks"""
    # input sky subtracted image
    ny,nx = im.shape
    nbin = 3  # 5
    nhbin = nbin//2

    mnim = np.zeros(im.shape,float)-100000
    
    count = 0
    xpeak = np.zeros(100000,float)
    ypeak = np.zeros(100000,float)
    for i in np.arange(nhbin+1,nx-nhbin-2):
        for j in range(nhbin+1,ny-nhbin-2):
            if im[j,i]>nsig*sig:
                if mnim[j,i] > -1000:
                    mval = mnim[j,i]
                else:
                    mval = np.mean(im[j-nhbin:j+nhbin+1,i-nhbin:i+nhbin+1])
                    mnim[j,i] = mval
                if mnim[j,i-1] > -1000:
                    lval = mnim[j,i-1]
                else:
                    lval = np.mean(im[j-nhbin:j+nhbin+1,i-nhbin-1:i+nhbin])
                    mnim[j,i-1] = lval
                if mnim[j,i+1] > -1000:
                    rval = mnim[j,i+1]
                else:
                    rval = np.mean(im[j-nhbin:j+nhbin+1,i-nhbin+1:i+nhbin+2])
                    mnim[j,i+1] = rval
                if mnim[j-1,i] > -1000:
                    dval = mnim[j-1,i]
                else:
                    dval = np.mean(im[j-nhbin-1:j+nhbin,i-nhbin:i+nhbin+1])
                    mnim[j-1,i] = dval
                if mnim[j+1,i] > -1000:
                    uval = mnim[j+1,i]
                else:
                    uval = np.mean(im[j-nhbin+1:j+nhbin+2,i-nhbin:i+nhbin+1])
                    mnim[j+1,i] = uval
                # Check that it is a peak
                if (mval>lval and mval>rval and mval>dval and mval>uval):
                    xpeak[count] = i
                    ypeak[count] = j
                    count = count + 1
    return xpeak,ypeak,count

@njit
@cc.export('boundingbox', '(f8[:,:],i8,i8,f8,f8)')
def boundingbox(im,xp,yp,thresh,bmax):
    """ Get bounding box for the source """

    ny,nx = im.shape
    nbin = 3
    nhbin = nbin//2
    
    # Step left until you reach the threshold
    y0 = yp-nhbin
    if y0<0: y0=0
    y1 = yp+nhbin+1
    if y1>ny: y1=ny
    flag = False
    midval = np.mean(im[y0:y1,xp])
    count = 1
    while (flag==False):
        newval = np.mean(im[y0:y1,xp-count])
        if newval < thresh*midval or xp-count==0 or count==bmax:
            flag = True
        lastval = newval
        count += 1
    leftxp = xp-count+1
    # Step right until you reach the threshold
    flag = False
    count = 1
    while (flag==False):
        newval = np.mean(im[y0:y1,xp+count])
        if newval < thresh*midval or xp+count==(nx-1) or count==bmax:
            flag = True
        lastval = newval
        count += 1
    rightxp = xp+count-1
    # Step down until you reach the threshold
    x0 = xp-nhbin
    if x0<0: x0=0
    x1 = xp+nhbin+1
    if x1>nx: x1=nx
    flag = False
    midval = np.mean(im[yp,x0:x1])
    count = 1
    while (flag==False):
        newval = np.mean(im[yp-count,x0:x1])
        if newval < thresh*midval or yp-count==0 or count==bmax:
            flag = True
        lastval = newval
        count += 1
    downyp = yp-count+1
    # Step up until you reach the threshold
    flag = False
    count = 1
    while (flag==False):
        newval = np.mean(im[yp+count,x0:x1])
        if newval < thresh*midval or yp+count==(ny-1) or count==bmax:
            flag = True
        lastval = newval
        count += 1
    upyp = yp+count-1

    return leftxp,rightxp,downyp,upyp

@njit
@cc.export('morpho', '(f8[:,:],i8,i8,i8,i8,i8,i8,f8)')
def morpho(im,xp,yp,x0,x1,y0,y1,thresh):
    """ Measure morphology parameters """
    ny,nx = im.shape

    midval = im[yp,xp]
    hthresh = thresh*midval
    
    nx = x1-x0+1
    ny = y1-y0+1

    # Flux and first moments
    flux = 0.0
    mnx = 0.0
    mny = 0.0
    # X loop    
    for i in range(nx):
        x = i+x0
        # Y loop
        for j in range(ny):
            y = j+y0
            val = im[y,x]
            if val>hthresh:
                # Flux
                flux += val
                # First moments
                mnx += val*x
                mny += val*y
    mnx /= flux
    mny /= flux

    # Second moments
    sigx2 = 0.0
    sigy2 = 0.0
    sigxy = 0.0
    # X loop    
    for i in range(nx):
        x = i+x0
        # Y loop
        for j in range(ny):
            y = j+y0
            val = im[y,x]
            if val>hthresh:
                sigx2 += val*(x-mnx)**2
                sigy2 += val*(y-mny)**2
                sigxy += val*(x-mnx)*(y-mny)
    sigx2 /= flux
    sigy2 /= flux
    sigx = np.sqrt(sigx2)
    sigy = np.sqrt(sigy2)
    sigxy /= flux
    fwhm = (sigx+sigy)*0.5 * 2.35
    
    # Ellipse parameters
    asemi = np.sqrt( 0.5*(sigx2+sigy2) + np.sqrt(((sigx2-sigy2)*0.5)**2 + sigxy**2 ) )
    bsemi = np.sqrt( 0.5*(sigx2+sigy2) - np.sqrt(((sigx2-sigy2)*0.5)**2 + sigxy**2 ) )
    theta = 0.5*np.arctan2(2*sigxy,sigx2-sigy2)  # in radians

    return flux,mnx,mny,sigx,sigy,sigxy,fwhm,asemi,bsemi,theta

@njit
@cc.export('morphology', '(f8[:,:],f8[:],f8[:],f8,f8)')
def morphology(im,xpeak,ypeak,thresh,bmax):
    """ Measure morphology of the peaks."""

    ny,nx = im.shape
    nbin = 3
    nhbin = nbin//2

    mout = np.zeros((len(xpeak),17),float)
    for i in range(len(xpeak)):
        xp = int(xpeak[i])
        yp = int(ypeak[i])
        mout[i,0] = xp
        mout[i,1] = yp
        
        # Get the bounding box
        leftxp,rightxp,downyp,upyp = boundingbox(im,xp,yp,thresh,bmax)
        mout[i,2] = leftxp
        mout[i,3] = rightxp
        mout[i,4] = downyp
        mout[i,5] = upyp
        mout[i,6] = (rightxp-leftxp+1)*(upyp-downyp+1)


        # Measure morphology parameters
        out = morpho(im,xp,yp,leftxp,rightxp,downyp,upyp,thresh)
        #flux,mnx,mny,sigx,sigy,sigxy,fwhm,asemi,bsemi,theta = out
        mout[i,7:] = out

    return mout

# @njit
# def create_table(names, data):
#     """Creates a table-like structure using NumPy structured arrays."""
#     dtype = np.dtype([(name, data[0].dtype) for name in names])
#     table = np.zeros(len(data[0]), dtype=dtype)
#     for i, name in enumerate(names):
#         table[name] = data[i]
#     return table

# @njit
# def convertstringlist(norm_list):
#     numba_list = typed.List.empty_list(types.string)
#     for e in norm_list:
#         numba_list.append(e)
#     return numba_list

# @njit
# @cc.export('asciicode', 'i4(unicode_type)')
# def asciicode(a):
#     """ Get ascii code for a character."""
#     if a==',':
#         return 44
#     elif a=='.':
#         return 46
#     elif a=='0':
#         return 48
#     elif a=='1':
#         return 49
#     elif a=='2':
#         return 50
#     elif a=='3':
#         return 51
#     elif a=='4':
#         return 52
#     elif a=='5':
#         return 53
#     elif a=='6':
#         return 54
#     elif a=='7':
#         return 55
#     elif a=='8':
#         return 56
#     elif a=='9':
#         return 57
#     elif a=='A':
#         return 65
#     elif a=='B':
#         return 66
#     elif a=='C':
#         return 67
#     elif a=='D':
#         return 68
#     elif a=='E':
#         return 69
#     elif a=='F':
#         return 70
#     elif a=='G':
#         return 71
#     elif a=='H':
#         return 72
#     elif a=='I':
#         return 73
#     elif a=='J':
#         return 74
#     elif a=='K':
#         return 75
#     elif a=='L':
#         return 76
#     elif a=='M':
#         return 77
#     elif a=='N':
#         return 78
#     elif a=='O':
#         return 79
#     elif a=='P':
#         return 80
#     elif a=='Q':
#         return 81
#     elif a=='R':
#         return 82
#     elif a=='S':
#         return 83
#     elif a=='T':
#         return 84
#     elif a=='U':
#         return 85
#     elif a=='V':
#         return 86
#     elif a=='W':
#         return 87
#     elif a=='X':
#         return 88
#     elif a=='Y':
#         return 89
#     elif a=='Z':
#         return 90
#     elif a=='_':
#         return 95
#     elif a=='a':
#         return 97
#     elif a=='b':
#         return 98
#     elif a=='c':
#         return 99
#     elif a=='d':
#         return 100
#     elif a=='e':
#         return 101
#     elif a=='f':
#         return 102
#     elif a=='g':
#         return 103
#     elif a=='h':
#         return 104
#     elif a=='i':
#         return 105
#     elif a=='j':
#         return 106
#     elif a=='k':
#         return 107
#     elif a=='l':
#         return 108
#     elif a=='m':
#         return 109
#     elif a=='n':
#         return 110
#     elif a=='o':
#         return 111
#     elif a=='p':
#         return 112
#     elif a=='q':
#         return 113
#     elif a=='r':
#         return 114
#     elif a=='s':
#         return 115
#     elif a=='t':
#         return 116
#     elif a=='u':
#         return 117
#     elif a=='v':
#         return 118
#     elif a=='w':
#         return 119
#     elif a=='x':
#         return 120
#     elif a=='y':
#         return 121
#     elif a=='z':
#         return 122

# @njit
# @cc.export('asciicar', 'unicode_type(i4)')
# def asciichar(code):
#     """ Return ascii character given the code."""
#     vals = ['','','','','','','','','','',   # 0-9
#             '','','','','','','','','','',   # 10-19
#             '','','','','','','','','','',   # 20-29
#             '','','','','','','','','','',   # 30-39
#             '','','','',',','','.','','0','1',   # 40-49
#             '2','3','4','5','6','7','8','9','','',   # 50-59
#             '','','','','','A','B','C','D','E',   # 60-69
#             'F','G','H','I','J','K','L','M','N','O',   # 70-79
#             'P','Q','R','S','T','U','V','W','X','Y',   # 80-89
#             'Z','','','','','_','','a','b','c',   # 90-99
#             'd','e','f','g','h','i','j','k','l','m',   # 100-109
#             'n','o','p','q','r','s','t','u','v','w',   # 110-119
#             'x','y','z','','','','','','','']   # 120-129
#     return vals[code]
    
# @njit
# @cc.export('convertasciitoint', 'i8[:](unicode_type)')
# def convertasciitoint(data):
#     """ Convert ascii string to integer."""
#     n = len(data)
#     out = np.zeros(n,np.int64)
#     for i in range(n):
#         out[i] = asciicode(data[i])
#     return out

# @njit
# @cc.export('convertinttoascii', 'unicode_type(f8[:])')
# def convertinttoascii(data):
#     """ Convert array of ascii code integers to string."""
#     n = len(data)
#     out = ''
#     for i in range(n):
#         out += asciichar(data[i])
#     return out

# @njit
# @cc.export('asciiintvalue', 'i8(i4[:])')
# def asciiintvalue(arr):
#     """ Convert ascii code array to an integer value."""
#     out = 0
#     for i in range(len(arr)):
#         out += arr[i]*1000**i
#     return out

# @njit
# @cc.export('isinteger', 'b1(unicode_type)')
# def isinteger(val):
#     """ Check if this is an integer."""
#     sval = str(val)
#     sval1 = sval[0]
#     if (sval1=='1' or sval1=='2' or sval1=='3' or sval1=='4' or sval1=='5' or
#         sval1=='6' or sval1=='7' or sval1=='8' or sval1=='9'):
#         return True
#     else:
#         return False

# @njit
# @cc.export('formatfloat', 'unicode_type(f8,i4)')
# def formatfloat(val,ndigits):
#     """ Format a floating point number for string output."""
#     if val<0:
#         isnegative = True
#         ndig = ndigits-1
#         val = np.abs(val)
#     else:
#         isnegative = False
#         ndig = ndigits
#     exponent = int(np.log10(val))
#     if np.log10(val)<0:
#         exponent = int(np.log10(val))-1
#     significand = val/10.0**exponent
#     sigvals = str(int(significand*10.0**ndig))
#     osigvals = sigvals[0]+'.'+sigvals[1:]
#     # Whole value, >=1
#     if exponent < ndig and exponent >= 0:
#         sval = str(int(val*10.0**ndig))
#         oval = sval[:exponent+1]+'.'+sval[exponent+1:]
#         out = oval[:ndig]
#     # Whole value, <1
#     elif np.abs(exponent)<ndig and exponent<0:
#         sval = str(int(val*10.0**ndig))
#         aexponent = np.abs(exponent)
#         if aexponent > 1:
#             oval = '0.' + (aexponent-1)*'0' + sval
#         else:
#             oval = '0.'+sval
#         out = oval[:ndig]
#     # Scientific notation
#     else:
#         eout = 'e'+str(exponent)
#         out = osigvals[:ndig-len(eout)]+eout
#     if isnegative:
#         out = '-'+out
#     return out
    
# spec = [
#     #('_names', types.List(types.string)),
#     ('_namesintarray', types.int64[:,:]),
#     ('_namesnchars', types.int64[:]),
#     ('_namesintvalue', types.int64[:]),
#     ('ncols', types.int64),
#     ('nrows', types.int64),
#     ('data', types.float64[:,:]),
# ]
# @jitclass(spec)
# class Table(object):

#     def __init__(self, names, data):
#         n = len(names)
#         namesnchars = np.zeros(n,np.int64)
#         maxlength = 0
#         for i in range(n):
#             namesnchars[i] = len(names[i])
#             maxlength = np.max(np.array([maxlength,len(names[i])]))
#         self._namesnchars = namesnchars
#         namesintarray = np.zeros((n,maxlength),np.int64)-1
#         self._namesintarray = namesintarray
#         namesintvalue = np.zeros(n,np.int64)
#         self._namesintvalue = namesintvalue
#         for i in range(n):
#             nameint = convertasciitoint(names[i])
#             self._namesintarray[i,:len(nameint)] = nameint
#             self._namesintvalue[i] = asciiintvalue(nameint)
#         self.ncols = n
#         ncols,nrows = data.shape
#         self.nrows = nrows
#         self.data = np.zeros(data.shape,np.float64)
#         self.data[:,:] = data.astype(np.float64)

#     def __len__(self):
#         """ Return the number of rows."""
#         return self.nrows

#     def __str__(self):
#         """ string representation."""
#         # Header
#         out = ''
#         for i in range(self.ncols):
#             name1 = self.name(i)
#             n1 = len(name1)
#             if n1<11:
#                 out += name1+(11-n1)*' '
#             else:
#                 out += name1
#         out += '\n'
#         # Rows
#         for i in range(self.nrows):
#             out1 = ''
#             for j in range(self.ncols):
#                 out1 += formatfloat(self.data[j,i],10)+' '
#             out1 += '\n'
#             out += out1
#         return out

#     def name(self,i):
#         """ Return single name."""
#         nchar = self._namesnchars[i]
#         nameint = self._namesintarray[i,:]
#         nameint = nameint[:nchar]
#         colname = convertinttoascii(nameint)
#         return colname

#     @property
#     def names(self):
#         ls = typed.List.empty_list(types.string)
#         for i in range(self.ncols):
#             colname = self.name(i)
#             ls.append(colname)
#         return ls

#     def _getnameindex(self,name):
#         nameint = convertasciitoint(name)
#         nameval =  asciiintvalue(nameint)
#         ind, = np.where(self._namesintvalue==nameval)
#         return ind

#     def getcol(self,item):
#         data1 = np.zeros((self.ncols,1),np.float64)
#         data1[:,0] = self.data[:,item]
#         return Table(self.names,data1)
        
#     def __getitem__(self,item):
#         """ Get a single column."""
#         ind = self._getnameindex(item)
#         if len(ind)==0:
#             print('No column "'+str(item)+'"')
#             return
#         return self.data[ind,:]
        
#     def __setitem__(self,item,data):
#         """ Set the entire column."""
#         ind = self._getnameindex(item)
#         if len(ind)==0:
#             print('No column "'+str(item)+'"')
#             return
#         if len(data)!=self.nrows:
#             print(len(data),'elements input but need',self.nrows,'elements to set entire column')
#             return
#         self.data[ind[0],:] = data

#     def set(self,name,row,data):
#         """ Set a single value."""
#         ind = self._getnameindex(name)
#         if len(ind)==0:
#             print('No column "'+str(name)+'"')
#             return
#         if row>self.nrows-1:
#             print(row,'out of bounds for',self.nrows,'rows')
#         self.data[ind[0],row] = data

#     def setmany(self,name,rows,data):
#         """ Set multiple values in a single row."""
#         # rows and data must be numpy arrays
#         ind = self._getnameindex(name)
#         if len(ind)==0:
#             print('No column "'+str(name)+'"')
#             return
#         if len(rows) != len(data):
#             print('rows and data not the same length')
#             return
#         for i in range(len(rows)):
#             if rows[i]>self.nrows-1:
#                 print(rows[i],'out of bounds for',self.nrows,'rows')
#                 continue
#             self.data[ind[0],rows[i]] = data[i]


@njit
@cc.export('starbbox', 'i8[:](UniTuple(f8,2),UniTuple(i8,2),f8)')
def starbbox(coords,imshape,radius):
    """                                                                                         
    Return the boundary box for a star given radius and image size.                             
                                                                                                
    Parameters                                                                                  
    ----------                                                                                  
    coords: list or tuple                                                                       
       Central coordinates (xcen,ycen) of star (*absolute* values).                             
    imshape: list or tuple                                                                      
       Image shape (ny,nx) values.  Python images are (Y,X).                                    
    radius: float                                                                               
       Radius in pixels.                                                                        
                                                                                                
    Returns                                                                                     
    -------                                                                                     
    bbox : BoundingBox object                                                                   
       Bounding box of the x/y ranges.                                                          
       Upper values are EXCLUSIVE following the python convention.                              
                                                                                                
    """

    # pixels span +/-0.5 pixels in each direction
    # so pixel x=5 spans x=4.5-5.5
    # Use round() to get the value of the pixel that is covers
    # the coordinate
    # to include pixels at the bottom but not the top we have to
    # subtract a tiny bit before we round
    eta = 1e-5

    # Star coordinates
    xcen,ycen = coords
    ny,nx = imshape   # python images are (Y,X)
    x0 = xcen-radius
    x1 = xcen+radius
    y0 = ycen-radius
    y1 = ycen+radius
    xlo = np.maximum(int(np.round(x0-eta)),0)
    xhi = np.minimum(int(np.round(x1-eta))+1,nx)
    ylo = np.maximum(int(np.round(y0-eta)),0)
    yhi = np.minimum(int(np.round(y1-eta))+1,ny)
    # add 1 at the upper end because those values are EXCLUDED
    # by the standard python convention

    # The old way of doing it
    #xlo = np.maximum(int(np.floor(xcen-radius)),0)
    #xhi = np.minimum(int(np.ceil(xcen+radius+1)),nx)
    #ylo = np.maximum(int(np.floor(ycen-radius)),0)
    #yhi = np.minimum(int(np.ceil(ycen+radius+1)),ny)
    return np.array([xlo,xhi,ylo,yhi])


@njit
@cc.export('getstar', '(UniTuple(i8,2),f8,f8,i8,f8)')
def getstar(imshape,xcen,ycen,hpsfnpix,fitradius):
    """ Return a star's full footprint and fitted pixels data."""
    # always return the same size
    # a distance of 6.2 pixels spans 6 full pixels but you could have
    # 0.1 left on one side and 0.1 left on the other side
    # that's why we have to add 2 pixels
    nfpix = hpsfnpix*2+1
    fbbox = starbbox((xcen,ycen),imshape,hpsfnpix)
    nfx = fbbox[1]-fbbox[0]
    nfy = fbbox[3]-fbbox[2]
    # extra buffer is ALWAYS at the end of each dimension    
    fxdata = np.zeros(nfpix*nfpix,np.int64)-1
    fydata = np.zeros(nfpix*nfpix,np.int64)-1
    fravelindex = np.zeros(nfpix*nfpix,np.int64)-1
    fcount = 0
    # Fitting pixels
    npix = int(np.floor(2*fitradius))+2
    bbox = starbbox((xcen,ycen),imshape,fitradius)
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    xdata = np.zeros(npix*npix,np.int64)-1
    ydata = np.zeros(npix*npix,np.int64)-1
    ravelindex = np.zeros(npix*npix,np.int64)-1
    mask = np.zeros(npix*npix,np.int8)
    fcount = 0
    count = 0
    for j in range(nfpix):
        y = j + fbbox[2]
        for i in range(nfpix):
            x = i + fbbox[0]
            r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
            if x>=fbbox[0] and x<=fbbox[1]-1 and y>=fbbox[2] and y<=fbbox[3]-1:
                if r <= 1.0*hpsfnpix:
                    fxdata[fcount] = x
                    fydata[fcount] = y
                    fmulti_index = (np.array([y]),np.array([x]))
                    fravelindex[fcount] = ravel_multi_index(fmulti_index,imshape)[0]
                    fcount += 1
                if r <= fitradius:
                    xdata[count] = x
                    ydata[count] = y
                    multi_index = (np.array([y]),np.array([x]))
                    ravelindex[count] = ravel_multi_index(multi_index,imshape)[0]
                    mask[count] = 1
                    count += 1
    return (fxdata,fydata,fravelindex,fbbox,nfx,nfy,fcount,
            xdata,ydata,ravelindex,bbox,count,mask)

@njit
@cc.export('collatestars', '(UniTuple(i8,2),f8[:],f8[:],i8,f8)')
def collatestars(imshape,starx,stary,hpsfnpix,fitradius):
    """ Get full footprint and fitted pixels data for all stars."""
    nstars = len(starx)
    nfpix = 2*hpsfnpix+1
    npix = int(np.floor(2*fitradius))+2
    # Full footprint arrays
    fxdata = np.zeros((nstars,nfpix*nfpix),np.int64)
    fydata = np.zeros((nstars,nfpix*nfpix),np.int64)
    fravelindex = np.zeros((nstars,nfpix*nfpix),np.int64)
    fbbox = np.zeros((nstars,4),np.int32)
    fshape = np.zeros((nstars,2),np.int32)
    fndata = np.zeros(nstars,np.int32)
    # Fitting pixel arrays
    xdata = np.zeros((nstars,npix*npix),np.int64)
    ydata = np.zeros((nstars,npix*npix),np.int64)
    ravelindex = np.zeros((nstars,npix*npix),np.int64)
    bbox = np.zeros((nstars,4),np.int32)
    shape = np.zeros((nstars,2),np.int32)
    ndata = np.zeros(nstars,np.int32)
    mask = np.zeros((nstars,npix*npix),np.int8)
    for i in range(nstars):
        out = getstar(imshape,starx[i],stary[i],hpsfnpix,fitradius)
        # full footprint information
        fxdata1,fydata1,fravelindex1,fbbox1,fnx1,fny1,fn1 = out[:7]
        fxdata[i,:] = fxdata1
        fydata[i,:] = fydata1
        fravelindex[i,:] = fravelindex1
        fbbox[i,:] = fbbox1
        fshape[i,0] = fny1
        fshape[i,1] = fnx1
        fndata[i] = fn1
        # fitting pixel information
        xdata1,ydata1,ravelindex1,bbox1,n1,mask1 = out[7:]
        xdata[i,:] = xdata1
        ydata[i,:] = ydata1
        ravelindex[i,:] = ravelindex1
        bbox[i,:] = bbox1
        ndata[i] = n1
        mask[i,:] = mask1
    # Trim arrays
    maxfn = np.max(fndata)
    fxdata = fxdata[:,:maxfn]
    fydata = fydata[:,:maxfn]
    fravelindex = fravelindex[:,:maxfn]
    maxn = np.max(ndata)
    xdata = xdata[:,:maxn]
    ydata = ydata[:,:maxn]
    ravelindex = ravelindex[:,:maxn]
    mask = mask[:,:maxn]
    
    return (fxdata,fydata,fravelindex,fbbox,fshape,fndata,
            xdata,ydata,ravelindex,bbox,ndata,mask)


@njit
@cc.export('getfullstar', '(UniTuple(i8,2),f8,f8,i8)')
def getfullstar(imshape,xcen,ycen,hpsfnpix):
    """ Return the entire footprint image/error/x/y arrays for one star."""
    # always return the same size
    # a distance of 6.2 pixels spans 6 full pixels but you could have
    # 0.1 left on one side and 0.1 left on the other side
    # that's why we have to add 2 pixels
    npix = hpsfnpix*2+1
    bbox = starbbox((xcen,ycen),imshape,hpsfnpix)
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    # extra buffer is ALWAYS at the end of each dimension    
    xdata = np.zeros(npix*npix,np.int32)-1
    ydata = np.zeros(npix*npix,np.int32)-1
    ravelindex = np.zeros(npix*npix,np.int64)-1
    count = 0
    for j in range(npix):
        y = j + bbox[2]
        for i in range(npix):
            x = i + bbox[0]
            if x>=bbox[0] and x<=bbox[1]-1 and y>=bbox[2] and y<=bbox[3]-1:
                xdata[count] = x
                ydata[count] = y
                multi_index = (np.array([y]),np.array([x]))
                ravelindex[count] = ravel_multi_index(multi_index,imshape)[0]
                count += 1
    return xdata,ydata,ravelindex,bbox,nx,ny,count

@njit
@cc.export('collatefullstars', '(UniTuple(i8,2),f8[:],f8[:],i8)')
def collatefullstars(imshape,starx,stary,hpsfnpix):
    """ Get the entire footprint image/error/x/y for all of the stars."""
    nstars = len(starx)
    npix = 2*hpsfnpix+1
    # Get xdata, ydata, error
    xdata = np.zeros((nstars,npix*npix),np.int32)
    ydata = np.zeros((nstars,npix*npix),np.int32)
    ravelindex = np.zeros((nstars,npix*npix),np.int64)
    bbox = np.zeros((nstars,4),np.int32)
    shape = np.zeros((nstars,2),np.int32)
    ndata = np.zeros(nstars,np.int32)
    for i in range(nstars):
        xdata1,ydata1,ravelindex1,bbox1,nx1,ny1,n1 = getfullstar(imshape,starx[i],stary[i],hpsfnpix)
        xdata[i,:] = xdata1
        ydata[i,:] = ydata1
        ravelindex[i,:] = ravelindex1
        bbox[i,:] = bbox1
        shape[i,0] = ny1
        shape[i,1] = nx1
        ndata[i] = n1
    return xdata,ydata,ravelindex,bbox,shape,ndata

@njit
@cc.export('getfitstar', '(UniTuple(i8,2),f8,f8,f8)')
def getfitstar(imshape,xcen,ycen,fitradius):
    """ Get the fitting pixel information for a single star."""
    npix = int(np.floor(2*fitradius))+2
    bbox = starbbox((xcen,ycen),imshape,fitradius)
    nx = bbox[1]-bbox[0]
    ny = bbox[3]-bbox[2]
    xdata = np.zeros(npix*npix,np.int32)-1
    ydata = np.zeros(npix*npix,np.int32)-1
    ravelindex = np.zeros(npix*npix,np.int64)-1
    mask = np.zeros(npix*npix,np.int32)
    count = 0
    for j in range(ny):
        y = j + bbox[2]
        for i in range(nx):
            x = i + bbox[0]
            r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
            if r <= fitradius:
                xdata[count] = x
                ydata[count] = y
                multi_index = (np.array([y]),np.array([x]))
                ravelindex[count] = ravel_multi_index(multi_index,imshape)[0]
                mask[count] = 1
                count += 1
    return xdata,ydata,ravelindex,count,mask
        
@njit
@cc.export('collatefitstars', '(UniTuple(i8,2),f8[:],f8[:],f8)')
def collatefitstars(imshape,starx,stary,fitradius):
    """ Get the fitting pixel information for all stars."""
    nstars = len(starx)
    npix = int(np.floor(2*fitradius))+2
    # Get xdata, ydata, error
    maxpix = nstars*(npix)**2
    xdata = np.zeros((nstars,npix*npix),np.int32)
    ydata = np.zeros((nstars,npix*npix),np.int32)
    ravelindex = np.zeros((nstars,npix*npix),np.int64)
    mask = np.zeros((nstars,npix*npix),np.int32)
    ndata = np.zeros(nstars,np.int32)
    bbox = np.zeros((nstars,4),np.int32)
    for i in range(nstars):
        xcen = starx[i]
        ycen = stary[i]
        bb = starbbox((xcen,ycen),imshape,fitradius)
        xdata1,ydata1,ravelindex1,n1,mask1 = getfitstar(imshape,xcen,ycen,fitradius)
        xdata[i,:] = xdata1
        ydata[i,:] = ydata1
        ravelindex[i,:] = ravelindex1
        mask[i,:] = mask1
        ndata[i] = n1
        bbox[i,:] = bb
    return xdata,ydata,ravelindex,bbox,ndata,mask

@njit
@cc.export('skyval', 'f8(f8[:],f8[:])')
@cc.export('skyvali', 'f8(i8[:],f8[:])')
def skyval(array,sigma):
    """  Estimate sky value from sky pixels."""
    wt = 1/sigma**2
    med = np.median(array)
    sig = mad(array)
    # reweight outlier pixels using Stetson's method
    resid = array-med
    wt2 = wt/(1+np.abs(resid)**2/np.median(sigma))
    xmn = np.sum(wt2*array)/np.sum(wt2)
    return xmn

@njit
@cc.export('mkbounds', '(f8[:],UniTuple(i8,2),i8)')
def mkbounds(pars,imshape,xoff=10):
    """ Make bounds for a set of input parameters."""
    # is [amp1,xcen1,ycen1,amp2,xcen2,ycen2, ...]
    npars = len(pars)
    nstars = npars // 3
    skyfit = (npars % 3 != 0)
    ny,nx = imshape
    # Make bounds
    lbounds = np.zeros(npars,float)
    ubounds = np.zeros(npars,float)
    lbounds[0:3*nstars:3] = 0
    lbounds[1:3*nstars:3] = np.maximum(pars[1:3*nstars:3]-xoff,0)
    lbounds[2:3*nstars:3] = np.maximum(pars[2:3*nstars:3]-xoff,0)
    if skyfit:
        lbounds[-1] = -np.inf
    ubounds[0:3*nstars:3] = np.inf
    ubounds[1:3*nstars:3] = np.minimum(pars[1:3*nstars:3]+xoff,nx-1)
    ubounds[2:3*nstars:3] = np.minimum(pars[2:3*nstars:3]+xoff,ny-1)
    if skyfit:
        ubounds[-1] = np.inf
    bounds = (lbounds,ubounds)
    return bounds

@njit
@cc.export('steps', 'f8[:](f8[:],f8)')
@cc.export('stepsi', 'f8[:](i8[:],f8)')
def steps(pars,dx=0.5):
    """ Return step sizes to use when fitting the stellar parameters."""
    npars = len(pars)
    nstars = npars // 3
    skyfit = (npars % 3 != 0)
    fsteps = np.zeros(npars,float)
    fsteps[0:3*nstars:3] = np.maximum(np.abs(pars[0:3*nstars:3])*0.5,1)
    fsteps[1:3*nstars:3] = dx        
    fsteps[2:3*nstars:3] = dx
    if skyfit:
        fsteps[-1] = 10
    return fsteps

@njit
@cc.export('newpars', 'f8[:](f8[:],f8[:],UniTuple(f8[:],2),f8[:])')
def newpars(pars,steps,bounds,maxsteps):
    """ Get new parameters given initial parameters, steps and constraints."""
    # Limit the steps to maxsteps
    limited_steps = limsteps(steps,maxsteps)
    # Make sure that these don't cross the boundaries
    lbounds,ubounds = bounds
    check = checkbounds(pars+limited_steps,bounds)
    # Reduce step size for any parameters to go beyond the boundaries
    badpars = (check!=0)
    # reduce the step sizes until they are within bounds
    newsteps = limited_steps.copy()
    count = 0
    maxiter = 2
    while (np.sum(badpars)>0 and count<=maxiter):
        newsteps[badpars] /= 2
        newcheck = checkbounds(pars+newsteps,bounds)
        badpars = (newcheck!=0)
        count += 1
            
    # Final parameters
    newpars = pars + newsteps
            
    # Make sure to limit them to the boundaries
    check = checkbounds(newpars,bounds)
    badpars = (check!=0)
    if np.sum(badpars)>0:
        # add a tiny offset so it doesn't fit right on the boundary
        newpars = np.minimum(np.maximum(newpars,lbounds+1e-30),ubounds-1e-30)
    return newpars

@njit
@cc.export('doPolygonsOverlap', '(f8[:],f8[:],f8[:],f8[:])')
@cc.export('doPolygonsOverlapi', '(i8[:],i8[:],i8[:],i8[:])')
def doPolygonsOverlap(xPolygon1, yPolygon1, xPolygon2, yPolygon2):
    """Returns True if two polygons are overlapping."""

    # How to determine if two polygons overlap.
    # If a vertex of one of the polygons is inside the other polygon
    # then they overlap.
    
    n1 = len(xPolygon1)
    n2 = len(xPolygon2)
    isin = False

    # If ranges don't overlap, then polygons don't overlap
    if rangeoverlap(xPolygon1,xPolygon2)==False or rangeoverlap(yPolygon1,yPolygon2)==False:
        return False
    
    # Loop through all vertices of second polygon
    for i in range(n2):
        # perform iterative boolean OR
        # if any point is inside the polygon then they overlap   
        isin = isin or isPointInPolygon(xPolygon1, yPolygon1, xPolygon2[i], yPolygon2[i])

    # Need to do the reverse as well, not the same
    for i in range(n1):
        isin = isin or isPointInPolygon(xPolygon2, yPolygon2, xPolygon1[i], yPolygon1[i])

    # Two polygons can overlap even if there are no vertices inside each other.
    # Need to check if the line segments overlap
    if isin==False:
        intersect = False
        # Add first vertex to the end
        xp1 = np.append( xPolygon1, xPolygon1[0] )
        yp1 = np.append( yPolygon1, yPolygon1[0] )
        xp2 = np.append( xPolygon2, xPolygon2[0] )
        yp2 = np.append( yPolygon2, yPolygon2[0] )
        for i in range(4):
            for j in range(4):
                intersect = intersect or doLineSegmentsIntersect(xp1[i:i+2],yp1[i:i+2],xp2[j:j+2],yp2[j:j+2])
                if intersect==True:
                    return True
        isin = isin or intersect
        
    return isin

@njit
@cc.export('isPointInPolygon', '(f8[:],f8[:],f8,f8)')
@cc.export('isPointInPolygoni', '(i8[:],i8[:],i8,i8)')
def isPointInPolygon(xPolygon, yPolygon, xPt, yPt):
    """Returns boolean if a point is inside a polygon of vertices."""
    
    # How to tell if a point is inside a polygon:
    # Determine the change in angle made by the point and the vertices
    # of the polygon.  Add up the delta(angle)'s from the first (include
    # the first point again at the end).  If the point is inside the
    # polygon, then the total angle will be +/-360 deg.  If the point is
    # outside, then the total angle will be 0 deg.  Points on the edge will
    # outside.
    # This is called the Winding Algorithm
    # http://geomalgorithms.com/a03-_inclusion.html

    n = len(xPolygon)
    # Array for the angles
    angle = np.zeros(n)

    # add first vertex to the end
    xPolygon1 = np.append( xPolygon, xPolygon[0] )
    yPolygon1 = np.append( yPolygon, yPolygon[0] )

    wn = 0   # winding number counter

    # Loop through the edges of the polygon
    for i in range(n):
        # if edge crosses upward (includes its starting endpoint, and excludes its final endpoint)
        if yPolygon1[i] <= yPt and yPolygon1[i+1] > yPt:
            # if (P is  strictly left of E[i])    // Rule #4
            if isLeft(xPolygon1[i], yPolygon1[i], xPolygon1[i+1], yPolygon1[i+1], xPt, yPt) > 0: 
                 wn += 1   # a valid up intersect right of P.x

        # if edge crosses downward (excludes its starting endpoint, and includes its final endpoint)
        if yPolygon1[i] > yPt and yPolygon1[i+1] <= yPt:
            # if (P is  strictly right of E[i])    // Rule #4
            if isLeft(xPolygon1[i], yPolygon1[i], xPolygon1[i+1], yPolygon1[i+1], xPt, yPt) < 0: 
                 wn -= 1   # a valid up intersect right of P.x

    # wn = 0 only when P is outside the polygon
    if wn == 0:
        return False
    else:
        return True

@njit
@cc.export('isLeft', '(f8,f8,f8,f8,f8,f8)')
@cc.export('isLefti', '(i8,i8,i8,i8,i8,i8)')
def isLeft(x1, y1, x2, y2, x3, y3):
    # isLeft(): test if a point is Left|On|Right of an infinite 2D line.
    #   From http://geomalgorithms.com/a01-_area.html
    # Input:  three points P1, P2, and P3
    # Return: >0 for P3 left of the line through P1 to P2
    # =0 for P3 on the line
    # <0 for P3 right of the line
    return ( (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1) )

@njit
@cc.export('rangeoverlap', '(f8[:],f8[:])')
@cc.export('rangeoverlapi', '(i8[:],i8[:])')
def rangeoverlap(a,b):
    """does the range (start1, end1) overlap with (start2, end2)"""
    return max(a) >= min(b) and min(a) <= max(b)

@njit
@cc.export('doLineSegmentsIntersect', '(f8[:],f8[:],f8[:],f8[:])')
@cc.export('doLineSegmentsIntersecti', '(i8[:],i8[:],i8[:],i8[:])')
def doLineSegmentsIntersect(x1, y1, x2, y2):
    """ Do two line segments intersect."""

    # Check vertical lines

    # Vertical lines, but NOT same X-values
    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]!=x2[0]:
        return False  # No overlap
    
    # Vertical lines with same X values
    if x1[0]==x1[1] and x2[0]==x2[1] and x1[0]==x2[0]:
        # Check intersection of Y ranges
        I1 = [np.min(y1), np.max(y1)]
        I2 = [np.min(y2), np.max(y2)]
    
        # And we could say that Xa is included into :
        Ia = [max( np.min(y1), np.min(y2) ),
              min( np.max(y1), np.max(y2) )]
    
        # Now, we need to check that this interval Ia exists :
        if rangeoverlap(y1,y2)==False:
            return False  # There is no mutual abcisses        
        else:
            return True   # There is overlap

    # The equation of a line is:
    #
    # f(x) = A*x + b = y
    # For a segment, it is exactly the same, except that x is included on an interval I.
    # 
    # If you have two segments, defined as follow:
    #
    # Segment1 = {(X1, Y1), (X2, Y2)}
    # Segment2 = {(X3, Y3), (X4, Y4)}
    # The abcisse Xa of the potential point of intersection (Xa,Ya) must be contained in both interval I1 and I2, defined as follow:
    I1 = [np.min(x1), np.max(x1)]
    I2 = [np.min(x2), np.max(x2)]
    
    # And we could say that Xa is included into :
    Ia = [max( np.min(x1), np.min(x2) ),
          min( np.max(x1), np.max(x2) )]
    
    # Now, we need to check that this interval Ia exists :
    if rangeoverlap(x1,x2)==False:        
        return False  # There is no mutual abcisses

    # Check that the Y-ranges overlap as well
    if rangeoverlap(y1,y2)==False:        
        return False  # There is no mutual y-value overlap
    
    # So, we have two line formula, and a mutual interval. Your line formulas are:
    # f1(x) = m1*x + b1 = y
    # f2(x) = m2*x + b2 = y
    
    # As we got two points by segment, we are able to determine A1, A2, b1 and b2:
    dx1 = x1[1]-x1[0]
    if dx1==0:
        m1 = np.inf
        b1 = 0
    else:
        m1 = (y1[1]-y1[0])/(x1[1]-x1[0])  # Pay attention to not dividing by zero
        b1 = y1[0]-m1*x1[0]        
    dx2 = x2[1]-x2[0]
    if dx2==0:
        m2 = np.inf
        b2 = 0
    else:
        m2 = (y2[1]-y2[0])/(x2[1]-x2[0])  # Pay attention to not dividing by zero
        b2 = y2[0]-m2*x2[0]

    # If the segments are parallel, then m1 == m2:
    if (m1 == m2) and (b1 != b2):
        return False  # Parallel segments
    
    # If the segments are parallel and on top of each other, the m1==m2 and b1==b2
    # we've already required that the x-ranges (abcissas) overlap
    if (m1 == m2) and (b1 == b2):
        return True   # parallel segments on top of each other
    
    # A point (Xa,Ya) standing on both lines must satisfy both formulas f1 and f2:
    # Ya = m1 * Xa + b1
    # Ya = m2 * Xa + b2
    # A1 * Xa + b1 = m2 * Xa + b2

    # Line segment 1 is vertical line 
    if x1[0]==x1[1]:
        Xa = x1[0]
        Ya = m2*Xa+b2
        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):
            return True
        else:
            return False
    # Line semgent 2 is vertical line
    elif x2[0]==x2[1]:
        Xa = x2[0]
        Ya = m1*Xa+b1
        if rangeoverlap(x1,[Xa]) and rangeoverlap(x2,[Xa]) and \
           rangeoverlap(y1,[Ya]) and rangeoverlap(y2,[Ya]):        
            return True
        else:
            return False
    # Neither are vertical lines
    else:
        Xa = (b2 - b1) / (m1 - m2)
        
    # The last thing to do is check that Xa is included into Ia:
    if ( (Xa < max( np.min(x1), np.min(x2) )) or
         (Xa > min( np.max(x1), np.max(x2) )) ):
        return False  # intersection is out of bound
    else:
        return True




if __name__ == "__main__":
    cc.compile()



# cython: boundscheck=False
# cython: wraparound=False
# cython: language_level=3
# cython: cdivision=True
# cython: binding=False
# cython: inter_types=True
# cython: cpow=True

import cython
cimport cython
import numpy as np
cimport numpy as np
from cython.view cimport array as cvarray
from cpython cimport array
from scipy.special import gamma, gammaincinv, gammainc

from libc.math cimport exp,sqrt,atan2,pi,NAN,log,log10,abs,pow
from libcpp cimport bool
from libc.stdlib cimport malloc, free
from libc.stdio cimport printf
from cpython.mem cimport PyMem_Malloc, PyMem_Realloc, PyMem_Free

#cdef extern from "complex.h":
#    complex double cpow(complex double base, complex double exponent)

#include <time.h>

cdef extern from "math.h":
    double sin(double x)
    double cos(double x)
    #double atan2(double x)

cimport utils

from libc.time cimport time, time_t, clock, clock_t, CLOCKS_PER_SEC
from cpython.datetime cimport datetime


# import os
# import numpy as np
# from numba import njit,types,from_dtype
# from numba.experimental import jitclass
# from numba_kdtree import KDTree
# from . import models_numba as mnb, utils_numba as utils

# Fit a PSF model to multiple stars in an image

cdef double[:,:,:] starcube(double[:,:] tab,double[:,:] image, double[:,:] error,
                            int npix,double fillvalue):
    """
    Produce a cube of cutouts of stars.

    Parameters
    ----------
    tab : table
       The catalog of stars to use.  This should have "x" and "y" columns and
         preferably also "amp".
    image : CCDData object
       The image to use to generate the stellar images.
    fillvalue : float, optional
       The fill value to use for pixels that are bad are off the image.
            Default is np.nan.

    Returns
    -------
    cube : numpy array
       Two-dimensional cube (Npix,Npix,Nstars) of the star images.

    Example
    -------

    cube = starcube(tab,image)

    """
    cdef int nstars,nhpix
    cdef Py_ssize_t[:] imshape
    cdef double xcen,ycen
    cdef double[:] x,y
    cdef double[:,:] xx,yy,rr
    cdef double[:,:,:] cube

    imshape = image.shape
    # Get the residuals data
    nstars = len(tab)
    nhpix = npix//2
    cube = np.zeros((npix,npix,nstars),float)
    xx,yy = utils.meshgrid(np.arange(npix)-nhpix,np.arange(npix)-nhpix)
    rr = np.zeros((npix,npix),float)
    for i in range(npix):
        for j in range(npix):
            rr[i,j] = sqrt(xx[i,j]*xx[i,j]+yy[i,j]*yy[i,j])
    #rr = np.sqrt(xx*xx+yy*yy)        
    x = xx[0,:]
    y = yy[:,0]
    for i in range(nstars):
        #xcen = tab['x'][i]
        #ycen = tab['y'][i]
        xcen = tab[1,i]
        ycen = tab[2,i]
        bbox = utils.starbbox((xcen,ycen),imshape,nhpix)
        flux = utils.slice2d(image,bbox)
        err = utils.slice2d(error,bbox)
        #flux = bbox.slice(image)
        #err = bbox.slice(error)
        #amp = tab['peak'][i]
        amp = tab[0][i]
        xim,yim = bbox.xy()
        xim = xim-xcen
        yim = yim-ycen
        # Need to interpolate onto a regular grid
        im2 = np.zeros((npix,npix),float)+np.nan
        xind, = np.where((x>=bbox.ixmin-xcen) & (x<=bbox.ixmax-1-xcen))
        xmin,xmax = np.min(xind),np.max(xind)
        yind, = np.where((y>=bbox.iymin-ycen) & (y<=bbox.iymax-1-ycen))
        ymin,ymax = np.min(yind),np.max(yind)
        # xx/yy are relative coordinates for the final interpolated image
        #       i.e. [-3, -2, -1, 0, 1, 2, 3]
        # xim/yim are relative coordinates for the original flux image
        # for the interpolation we need coordinates relative to to (0,0) of the
        #    flux image
        fnx = xmax-xmin+1
        fny = ymax-ymin+1
        fxx = xx[ymin:ymax+1,xmin:xmax+1]-xim[0,0]   # relative to flux origin
        fyy = yy[ymin:ymax+1,xmin:xmax+1]-yim[0,0]
        fim = utils.alinearinterp(flux,fxx,fyy)
        fim = fim.reshape((fny,fnx))
        im2[ymin:ymax+1,xmin:xmax+1] = fim
        # Stuff it into 3D array
        cube[:,:,i] = im2
    return cube


cdef double[:,:,:] mkempirical(double[:,:,:] cube,int order,double[:,:] coords,Py_ssize_t[:] shape,int lookup):
    """
    Take a star cube and collapse it to make an empirical PSF using median
    and outlier rejection.

    Parameters
    ----------
    cube : numpy array
      Three-dimensional cube of star images (or residual images) of shape
        (Npix,Npix,Nstars).
    order : int
      The order of the variations. 0-constant, 1-linear terms.  If order=1,
        Then coords and shape must be input.
    coords : array
      Two-element tuple of the X/Y coordinates of the stars.  This is needed
        to generate the linear empirical model (order=1).
    shape : array
      Two-element tuple giving the shape (Ny,Nx) of the image.  This is
        needed to generate the linear empirical model (order=1).
    lookup : int,
      Parameter to indicate if this is a lookup table.  If lookup=False, then
      the constant term is constrained to be non-negative. Default is False.

    Returns
    -------
    epsf : numpy array
      The empirical PSF model  If order=0, then this is just a 2D image.  If
        order=1, then it will be a 3D cube (Npix,Npix,4) where the four terms
        are [constant, X-term, Y-term, X*Y-term].  If rect=True, then a list
        of RectBivariateSpline functions are returned.

    Example
    -------

    epsf = mkempirical(cube,order=0)

    or

    epsf = mkempirical(cube,order=1,coords=coords,shape=im.shape)

    """

    cdef Py_ssize_t nx,ny,nstar,npix,nhpix
    cdef int i
    cdef double gmed
    cdef Py_ssize_t[:] badx,bady,badk
    cdef double[:] error1
    cdef double[:,:] med,sig
    cdef double[:,:,:] epsf
    
    ny = cube.shape[0]
    nx = cube.shape[1]
    nstar = cube.shape[2]
    npix = ny
    nhpix = ny//2
    
    # Do outlier rejection in each pixel
    #med = utils.median3d(cube,axis=2,ignore_nan=True)
    med = np.median(cube,axis=2,ignore_nan=True)
    bady,badx = np.where(np.isfinite(med)==False)
    if len(bady)>0:
        gdmed = np.median(med,ignore_nan=True)
        for i in range(len(bady)):
            med[bady[i],badx[i]] = gdmed
    sig = utils.mad3d(cube,2,1,0)
    bady,badx = np.where(np.isfinite(sig)==False)
    if len(bady)>0:
        gdsig = np.median(sig,ignore_nan=True)
        for i in range(len(bady)):
            sig[bady[i],badx[i]] = gdsig
    # Mask outlier points
    bady,badx,badk = np.where((np.abs(cube-med.reshape((med.shape)+(-1,)))>3*sig.reshape((med.shape)+(-1,)))
                              & np.isfinite(cube))
    if len(bady)>0:
        macube = cube.copy()
        for i in range(len(bady)):
            macube[bady[i],badx[i],badk[i]] = np.nan
        medim = np.mean(macube,axis=2,ignore_nan=True)
        # Remove any final NaNs
        bady,badx = np.where(np.isfinite(medim)==False)
        if len(bady)>0:
            gdmed = np.median(medim,ignore_nan=True)
            for i in range(len(bady)):
                medim[bady[i],badx[i]] = gdmed
    else:
        medim = med

    # Check how well each star fits the median
    goodpix = np.sum(np.sum(np.isfinite(macube),axis=0),axis=0)
    print('goodpix=',goodpix)
    #goodpix = macube.count(axis=(0,1))
    resid2 = (cube-medim.reshape((medim.shape)+(-1,)))**2
    mnresid2 = np.sum(np.sum(resid2,axis=0,ignore_nan=True),axis=0,ignore_nan=True)/goodpix
    rms = np.sqrt(mnresid2)
    print('rms=',rms)
    #print(utils.sum3d(dum,axis=0,ignore_nan=True))
    #rms = np.sqrt(np.nansum((cube-medim.reshape((medim.shape)+(-1,)))**2,axis=(0,1))/goodpix)

    xx,yy = utils.meshgrid(np.arange(npix)-nhpix,np.arange(npix)-nhpix)
    rr = np.sqrt(xx**2+yy**2)        
    x = xx[0,:]
    y = yy[:,0]
    mask = (rr<=nhpix)
    
    # Constant
    if order==0:
        # Make sure it goes to zero at large radius
        medim *= mask  # mask corners
        # Make sure values are positive
        if lookup==0:
            medim = np.maximum(medim,0.0)
            fpars = medim
        else:
            fpars = medim
        fpars = fpars.reshape((ny,nx,1))
            
    # Linear
    elif order==1:
        if coords is None or shape is None:
            raise ValueError('Need coords and shape with order=1')
        pars = np.zeros((ny,nx,4),float)
        # scale coordinates to -1 to +1
        xcen,ycen = coords[:,0],coords[:,1]
        relx,rely = utils.relcoord(xcen,ycen,shape)
        # Loop over pixels and fit line to x/y
        for i in range(ny):
            for j in range(nx):
                data1 = cube[i,j,:]
                if np.sum(np.abs(data1)) != 0:
                    # maybe use a small maxiter
                    error1 = np.ones(len(data1))
                    pars1,perror1,pcov1 = utils.poly2dfit(relx,rely,data1,error1,2,0.5)
                    pars[i,j,:] = pars1
        # Make sure it goes to zero at large radius
        fpars = pars.copy()
        for i in range(4):
            # Make sure edges are zero on average for higher-order terms
            gdouter, = np.where(rr.ravel()>nhpix*0.8)
            medouter = np.median(pars[:,:,i].ravel()[gdouter])
            fpars[:,:,i] -= medouter
            # Mask corners
            fpars[:,:,i] *= mask
            # Each higher-order term must have ZERO total volume
   
    return fpars,rms


# cdef double[:] findpsfnei(double[:,:] alltab,double[:,:] psftab,int npix):
#     """
#     Find stars near PSF stars.

#     Parameters
#     ----------
#     alltab : numpy array
#       Catalog of all sources in the image.
#     psftab : numpy array
#       Catalog of PSF stars.
#     npix : int
#       Search radius in pixels.

#     Returns
#     -------
#     indall : numpy array
#        List of indices into allcat that are neighbors to
#         the PSF stars (within radius of npix), and are not
#         PSF stars themselves.

#     Example
#     -------

#     indall = findpsfnei(alltab,psftab,npix)

#     """
#     # Returns distance and index of closest neighbor
    
#     nalltab = len(alltab)
#     npsftab = len(psftab)
    
#     # Use KD-tree
#     x1 = np.zeros((nalltab,2),float)
#     x1[:,0] = alltab[:,2]
#     x1[:,1] = alltab[:,3]
#     x2 = np.zeros((npsftab,2),float)
#     x2[:,0] = psftab[:,2]
#     x2[:,1] = psftab[:,3]


#     #data = np.random.random(3_000_000).reshape(-1, 3)
#     #kdtree = KDTree(data, leafsize=10)

#     ## query the nearest neighbors of the first 100 points
#     #distances, indices, neighbors = kdtree.query(data[:100], k=30)

#     ## query all points in a radius around the first 100 points
#     #indices = kdtree.query_radius(data[:100], r=0.5, return_sorted=True)

    
#     #kdt = cKDTree(x1)
#     kdt = KDTree(x1)
#     # Get distance for 2 closest neighbors
#     dist, ind = kdt.query(x2, k=50, distance_upper_bound=np.sqrt(2)*npix//2)
#     # closest neighbor is always itself, remove it 
#     dist = dist[:,1:]
#     ind = ind[:,1:]
#     # Add up all the stars
#     gdall, = np.where(np.isfinite(dist.ravel()))
#     indall = ind.ravel()[gdall]
#     # Get unique ones
#     indall = np.unique(indall)

#     # Make sure none of them are our psf stars
#     ind1,ind2,dist = utils.xmatch(alltab[indall,2],alltab[indall3],
#                                   psftab[:,2],psftab[:,3],5,sphere=False)
#     #    alltab['x'][indall],alltab['y'][indall],psftab['x'],psftab['y'],5)
#     # Remove any matches
#     if len(ind1)>0:
#         indall = np.delete(indall,ind1)
    
#     return indall


# cdef double[:,:] subtractnei(double[:,:] image,double[:,:] error,double[:,:] alltab,double[:,:] psftab,double[:] psf):
#     """
#     Subtract neighboring stars to PSF stars from the image.

#     Parameters
#     ----------
#     image : CCDDdata object
#        The input image from which to subtract PSF neighbor stars.
#     alltab : numpy array
#       Catalog of all sources in the image.
#     psftab : numpy array
#       Catalog of PSF stars.
#     psf : PSF object
#       The PSF model.

#     Returns
#     -------
#     resid : CCDData object
#       The input images with the PSF neighbor stars subtracted.

#     Example
#     -------

#     subim = subtractnei(image,alltab,psftab,psf)

#     """

#     indnei = findpsfnei(alltab,psftab,psf.npix)
#     nnei = len(indnei)

#     flux = image
#     resid = image.copy()
#     fitradius = psf.fwhm()*0.5
    
#     # Loop over neighboring stars and fit just the core
#     for i in range(nnei):
#         x1 = alltab[indnei[i],2]
#         xp1 = int(np.minimum(np.maximum(np.round(x1),0),image.shape[1]-1))
#         y1 = alltab[indnei[i],3]
#         yp1 = int(np.minimum(np.maximum(np.round(y1),0),image.shape[0]-1))
#         h1 = alltab[indnei[i],1]
#         initpars = np.array([h1,x1,y1])
#         bbox = starbbox((initpars[1],initpars[2]),image.shape,psf.radius)
#         # Fit amp empirically with central pixels
#         flux1 = bbox.slice(flux)
#         err1 = bbox.slice(error)
#         model1 = psf(pars=initpars,bbox=bbox)
#         good = ((flux1/err1>2) & (flux1>0) & (model1/np.max(model1)>0.25))
#         amp = np.median(flux1[good]/model1[good]) * initpars[0]
#         pars = np.array([amp, x1, y1])
#         im1 = psf(pars=pars,bbox=bbox)
#         resid[bbox.slices].data -= im1
#     return resid

# 
# def sliceinsert(array,lo,insert):
#     """ Insert array values."""
#     n = insert.size
#     for i in range(n):
#         j = i + lo
#         array[j] = insert[i]
#     # array is updated in place  



# cdef list getstar(double[:,:] image,double[:,:] error,double xcen,double ycen,double fitradius):
#     """ Return the entire footprint image/error/x/y arrays for one star."""
#     # always return the same size
#     # a distance of 6.2 pixels spans 6 full pixels but you could have
#     # 0.1 left on one side and 0.1 left on the other side
#     # that's why we have to add 2 pixels
#     npix = int(np.floor(2*fitradius))+2
#     bbox = starbbox((xcen,ycen),image.shape,fitradius)
#     nx = bbox[1]-bbox[0]
#     ny = bbox[3]-bbox[2]
#     # extra buffer is ALWAYS at the end of each dimension    
#     imdata = np.zeros(npix*npix,float)+np.nan
#     errdata = np.zeros(npix*npix,float)+np.nan
#     xdata = np.zeros(npix*npix,np.int32)-1
#     ydata = np.zeros(npix*npix,np.int32)-1
#     count = 0
#     for j in range(npix):
#         y = j + bbox[2]
#         for i in range(npix):
#             x = i + bbox[0]            
#             xdata[count] = x
#             ydata[count] = y
#             if x>=bbox[0] and x<=bbox[1]-1 and y>=bbox[2] and y<=bbox[3]-1:
#                 imdata[count] = image[y,x]
#                 errdata[count] = error[y,x]
#             count += 1
#     return imdata,errdata,xdata,ydata,bbox,nx,ny


# cdef list collatestars(double[:,:] image,double[:,:] error,double[:] starx,double[:] stary,double fitradius):
#     """ Get the entire footprint image/error/x/y for all of the stars."""
#     nstars = len(starx)
#     npix = int(np.floor(2*fitradius))+2
#     # Get xdata, ydata, error
#     imdata = np.zeros((nstars,npix*npix),float)
#     errdata = np.zeros((nstars,npix*npix),float)
#     xdata = np.zeros((nstars,npix*npix),np.int32)
#     ydata = np.zeros((nstars,npix*npix),np.int32)
#     bbox = np.zeros((nstars,4),np.int32)
#     shape = np.zeros((nstars,2),np.int32)
#     for i in range(nstars):
#         imdata1,errdata1,xdata1,ydata1,bbox1,nx1,ny1 = getstar(image,error,starx[i],stary[i],fitradius)
#         imdata[i,:] = imdata1
#         errdata[i,:] = errdata1
#         xdata[i,:] = xdata1
#         ydata[i,:] = ydata1
#         bbox[i,:] = bbox1
#         shape[i,0] = ny1
#         shape[i,1] = nx1
#     return imdata,errdata,xdata,ydata,bbox,shape


# cdef list unpackstar(double[:,:] imdata,double[:,:] errdata,double[:] xdata,double[:] ydata,
#                      int[:] bbox,int[:] shape,int istar):
#     """ Return unpacked data for one star."""
#     imdata1 = imdata[istar,:].copy()
#     errdata1 = errdata[istar,:].copy()
#     xdata1 = xdata[istar,:].copy()
#     ydata1 = ydata[istar,:].copy()
#     bbox1 = bbox[istar,:]
#     shape1 = shape[istar,:]
#     n = len(imdata1)
#     npix = int(np.sqrt(n))
#     # Convert to 2D arrays
#     imdata1 = imdata1.reshape(npix,npix)
#     errdata1 = errdata1.reshape(npix,npix)
#     xdata1 = xdata1.reshape(npix,npix)
#     ydata1 = ydata1.reshape(npix,npix)
#     # Trim values
#     if shape1[0] < npix or shape1[1] < npix:
#         imdata1 = imdata1[:shape1[0],:shape1[1]]
#         errdata1 = errdata1[:shape1[0],:shape1[1]]
#         xdata1 = xdata1[:shape1[0],:shape1[1]]
#         ydata1 = ydata1[:shape1[0],:shape1[1]]
#     return imdata1,errdata1,xdata1,ydata1,bbox1,shape1


# cdef list getfitstar(double[:,:] image,double[:,:] error,double xcen,double ycen,double fitradius):
#     """ Get the fitting pixels for a single star."""
#     npix = int(np.floor(2*fitradius))+2
#     bbox = starbbox((xcen,ycen),image.shape,fitradius)
#     flux = image[bbox[2]:bbox[3],bbox[0]:bbox[1]].copy()
#     err = error[bbox[2]:bbox[3],bbox[0]:bbox[1]].copy()
#     nflux = flux.size
#     nx = bbox[1]-bbox[0]
#     ny = bbox[3]-bbox[2]
#     imdata = np.zeros(npix*npix,float)
#     errdata = np.ones(npix*npix,float)
#     xdata = np.zeros(npix*npix,np.int32)-1
#     ydata = np.zeros(npix*npix,np.int32)-1
#     mask = np.zeros(npix*npix,np.int32)
#     count = 0
#     for j in range(ny):
#         y = j + bbox[2]
#         for i in range(nx):
#             x = i + bbox[0]
#             r = np.sqrt((x-xcen)**2 + (y-ycen)**2)
#             if r <= fitradius:
#                 imdata[count] = flux[j,i]
#                 errdata[count] = error[j,i]
#                 xdata[count] = x
#                 ydata[count] = y
#                 mask[count] = 1
#                 count += 1
#     return imdata,errdata,xdata,ydata,count,mask
        

# cdef list collatefitstars(double[:,:] image,double[:,:] error,double[:] starx,double[:] stary,double fitradius):
#     nstars = len(starx)
#     npix = int(np.floor(2*fitradius))+2
#     # Get xdata, ydata, error
#     maxpix = nstars*(npix)**2
#     imdata = np.zeros((nstars,npix*npix),float)
#     errdata = np.zeros((nstars,npix*npix),float)
#     xdata = np.zeros((nstars,npix*npix),np.int32)
#     ydata = np.zeros((nstars,npix*npix),np.int32)
#     mask = np.zeros((nstars,npix*npix),np.int32)
#     ndata = np.zeros(nstars,np.int32)
#     bbox = np.zeros((nstars,4),np.int32)
#     for i in range(nstars):
#         xcen = starx[i]
#         ycen = stary[i]
#         bb = starbbox((xcen,ycen),image.shape,fitradius)
#         imdata1,errdata1,xdata1,ydata1,n1,mask1 = getfitstar(image,error,xcen,ycen,fitradius)
#         imdata[i,:] = imdata1
#         errdata[i,:] = errdata1
#         xdata[i,:] = xdata1
#         ydata[i,:] = ydata1
#         mask[i,:] = mask1
#         ndata[i] = n1
#         bbox[i,:] = bb
#     return imdata,errdata,xdata,ydata,bbox,ndata,mask


# cdef list unpackfitstar(double[:,:] imdata,double[:,:] errdata,double[:] xdata,double[:] ydata,
#                         int[:] bbox,int ndata,int istar):
#     """ Return unpacked fitting data for one star."""
#     imdata1 = imdata[istar,:]
#     errdata1 = errdata[istar,:]
#     xdata1 = xdata[istar,:]
#     ydata1 = ydata[istar,:]
#     bbox1 = bbox[istar,:]
#     n1 = ndata[istar]
#     # Trim to the values we want
#     imdata1 = imdata1[:n1]
#     errdata1 = errdata1[:n1]
#     xdata1 = xdata1[:n1]
#     ydata1 = ydata1[:n1]
#     return imdata1,errdata1,xdata1,ydata1,bbox1,n1

    
# kv_ty = (types.int64, types.unicode_type)
# spec = [
#     ('psftype', types.int32),
#     ('params', types.float64[:]),
#     #('lookup', types.float64[:,:,:]),
#     #('order', types.int32),
#     #('fwhm', types.float64),
#     ('image', types.float64[:,:]),
#     ('error', types.float64[:,:]),
#     ('starinit', types.float64[:,:]),
#     ('nstars', types.int32),
#     ('niter', types.int32),
#     ('npsfpix', types.int32),
#     ('nx', types.int32),
#     ('ny', types.int32),
#     ('fitradius', types.float64),
#     ('nfitpix', types.int32),
#     ('staramp', types.float64[:]),
#     ('starxcen', types.float64[:]),
#     ('starycen', types.float64[:]),
#     ('starchisq', types.float64[:]),
#     ('starrms', types.float64[:]),
#     ('starnpix', types.int32[:]),
#     ('star_imdata', types.float64[:,:]),
#     ('star_errdata', types.float64[:,:]),
#     ('star_xdata', types.int32[:,:]),
#     ('star_ydata', types.int32[:,:]),
#     ('star_bbox', types.int32[:,:]),
#     ('star_shape', types.int32[:,:]),
#     ('starfit_imdata', types.float64[:,:]),
#     ('starfit_errdata', types.float64[:,:]),
#     ('starfit_xdata', types.int32[:,:]),
#     ('starfit_ydata', types.int32[:,:]),
#     ('starfit_mask', types.int32[:,:]),
#     ('starfit_bbox', types.int32[:,:]),
#     ('starfit_ndata', types.int32[:]),
#     ('npix', types.int32),
#     ('radius', types.int32),
#     ('verbose', types.boolean),
#     ('niter', types.int32),
#     ('_unitfootflux', types.float64),
#     #('lookup', types.float64[:,:,:]),
#     #('_bounds', types.float64[:,:]),
#     #('_steps', types.float64[:]),
#     ('coords', types.float64[:]),
#     ('imshape', types.int32[:]),
#     #('order', types.int32),
# ]

# @jitclass(spec)
# class PSFFitter(object):

#     def __init__(self,psftype,psfparams,image,error,starx,stary,starflux,fitradius,verbose):
#         self.verbose = verbose
#         self.psftype = psftype
#         self.params = psfparams
#         #self.lookup = psf.lookup
#         #self.order = psf.order
#         #self.fwhm = psf.fwhm()
#         self.image = image.astype(np.float64)
#         self.error = error.astype(np.float64)
#         self.imshape = np.zeros(2,np.int32)
#         self.imshape[:] = image.shape
#         nstars = len(starx)
#         self.starinit = np.zeros((nstars,3),float)
#         self.starinit[:,0] = starx
#         self.starinit[:,1] = stary
#         self.starinit[:,2] = starflux
#         self.nstars = nstars
#         self.niter = 0
#         #self.npsfpix = psf.npix
#         ny,nx = image.shape
#         self.nx = nx
#         self.ny = ny
#         #if np.isfinite(fitradius)==False:
#         #    fitradius = self.fwhm*1.5
#         self.fitradius = fitradius
#         self.nfitpix = int(np.floor(2*fitradius))+2  # max pixels 
#         self.staramp = np.zeros(self.nstars,float)
#         self.staramp[:] = starflux/(2*fitradius+1)**2
#         # current best-fit values
#         self.starxcen = np.zeros(self.nstars,float)
#         self.starxcen[:] = self.starinit[:,0].copy()
#         self.starycen = np.zeros(self.nstars,float)
#         self.starycen[:] = self.starinit[:,1].copy()
#         self.starchisq = np.zeros(self.nstars,float)
#         self.starrms = np.zeros(self.nstars,float)
#         self.starnpix = np.zeros(self.nstars,np.int32)

#         # Get information for all the stars
#         imdata,errdata,xdata,ydata,bbox,shape = collatestars(image,error,self.starxcen,
#                                                              self.starycen,fitradius)
#         self.star_imdata = imdata
#         self.star_errdata = errdata
#         self.star_xdata = xdata
#         self.star_ydata = ydata
#         self.star_bbox = bbox
#         self.star_shape = shape
        
#         # Get fitting information for all the stars
#         fimdata,ferrdata,fxdata,fydata,fbbox,fndata,fmask = collatefitstars(image,error,self.starxcen,
#                                                                             self.starycen,fitradius)
#         self.starfit_imdata = fimdata
#         self.starfit_errdata = ferrdata
#         self.starfit_xdata = fxdata
#         self.starfit_ydata = fydata
#         self.starfit_mask = fmask
#         self.starfit_bbox = fbbox
#         self.starfit_ndata = fndata

#     def unpackstar(self,istar):
#         return unpackstar(self.star_imdata,self.star_errdata,self.star_xdata,
#                           self.star_ydata,self.star_bbox,self.star_shape,istar)

#     def unpackfitstar(self,istar):
#         return unpackfitstar(self.starfit_imdata,self.starfit_errdata,self.starfit_xdata,
#                              self.starfit_ydata,self.starfit_bbox,self.starfit_ndata,istar)

#     def psf(self,x,y,pars,psfparams,lookup,deriv):
#         """ Get a PSF model for a single star."""
#         return mnb.psf(x,y,pars,self.psftype,psfparams,lookup,
#                        self.imshape,deriv=deriv,verbose=self.verbose)

#     def model(self,x,args):
#         """ model function."""
#         # input the model parameters
        
#         if self.verbose:
#             print('model: '+str(self.niter)+' ',args)
#         nparsarr = np.array([6,7,8,8,7])
#         npars = nparsarr[self.psftype-1]
#         # Loop over the stars and generate the model image
#         allmodel = np.zeros(self.star_imdata.shape,float)
#         maxpix = self.star_imdata.shape[1]
#         for i in range(self.nstars):
#             im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
#             allpars = np.zeros(npars,float)
#             allpars[0] = self.staramp[i]
#             allpars[1] = self.starxcen[i]
#             allpars[2] = self.starycen[i]
#             allpars[3:] = args
#             g,_ = mnb.amodel2d(xdata1,ydata1,self.psftype,allpars,0)
#             allmodel[i,:len(g)] = g
#         return allmodel

#     def chisq(self,pars):
#         """ Create the model and return chisq for this set of PSF parameters."""
#         m = self.model(np.arange(10),pars)
#         chisq = np.sum((self.starfit_imdata-m)**2/self.starfit_errdata**2)        
#         return chisq
        
    
#             #image = self.imdata[i]
#             #amp = self.staramp[i]
#             #xcenorig = self.starxcenorig[i]   
#             #ycenorig = self.starycenorig[i]
#             #xcen = self.starxcen[i]   
#             #ycen = self.starycen[i]            
#             #bbox = self.bboxdata[i]
#             #x = self.xlist[i]
#             #y = self.ylist[i]
#             #pixstart = self.pixstart[i]
#             #npix = self.npix[i]
#             #flux = self.imflatten[pixstart:pixstart+npix]
#             #err = self.errflatten[pixstart:pixstart+npix]

#             #x0orig = xcenorig - bbox.ixmin
#             #y0orig = ycenorig - bbox.iymin
#             #x0 = xcen - bbox.ixmin
#             #y0 = ycen - bbox.iymin            
            
#             # # Fit amp/xcen/ycen if niter=1
#             # if refit:
#             #     #if (self.niter<=1): # or self.niter%3==0):
#             #     if self.niter>-1:
#             #         # force the positions to stay within +/-2 pixels of the original values
#             #         bounds = (np.array([0,np.maximum(x0orig-2,0),np.maximum(y0orig-2,0),-np.inf]),
#             #                   np.array([np.inf,np.minimum(x0orig+2,bbox.shape[1]-1),np.minimum(y0orig+2,bbox.shape[0]-1),np.inf]))
#             #         # the image still has sky in it, use sky (nosky=False)
#             #         if np.isfinite(psf.fwhm())==False:
#             #             print('nan fwhm')
#             #             #import pdb; pdb.set_trace()
#             #         pars,perror,model = psf.fit(image,[amp,x0,y0],nosky=False,retpararray=True,niter=5,bounds=bounds)
#             #         xcen += (pars[1]-x0)
#             #         ycen += (pars[2]-y0)
#             #         amp = pars[0]                    
#             #         self.staramp[i] = amp
#             #         self.starxcen[i] = xcen
#             #         self.starycen[i] = ycen
#             #         model = psf(x,y,pars=[amp,xcen,ycen])
#             #         if verbose:
#             #             print('Star '+str(i)+' Refitting all parameters')
#             #             print(str([amp,xcen,ycen]))

#             #         #pars2,model2,mpars2 = psf.fit(image,[amp,x0,y0],nosky=False,niter=5,allpars=True)
#             #         #import pdb; pdb.set_trace()
                        
#     #             # Only fit amp if niter>1
#     #             #   do it empirically
#     #             else:
#     #                 #im1 = psf(pars=[1.0,xcen,ycen],bbox=bbox)
#     #                 #wt = 1/image.error**2
#     #                 #amp = np.median(image.data[mask]/im1[mask])                
#     #                 model1 = psf(x,y,pars=[1.0,xcen,ycen])
#     #                 wt = 1/err**2
#     #                 amp = np.median(flux/model1)
#     #                 #amp = np.median(wt*flux/model1)/np.median(wt)

#     #                 self.staramp[i] = amp
#     #                 model = model1*amp
#     #                 #self.starxcen[i] = pars2[1]+xy[0][0]
#     #                 #self.starycen[i] = pars2[2]+xy[1][0]       
#     #                 #print(count,self.starxcen[i],self.starycen[i])
#     #                 # updating the X/Y values after the first iteration
#     #                 #  causes problems.  bounces around too much

#     #                 if verbose:
#     #                     print('Star '+str(i)+' Refitting amp empirically')
#     #                     print(str(amp))
                        
#     #                 #if i==1: print(amp)
#     #                 #if self.niter==2:
#     #                 #    import pdb; pdb.set_trace()

#     #         # No refit of stellar parameters
#     #         else:
#     #             model = psf(x,y,pars=[amp,xcen,ycen])

#     #         #if self.niter>1:
#     #         #    import pdb; pdb.set_trace()
                
#     #         # Relculate reduced chi squared
#     #         chisq = np.sum((flux-model.ravel())**2/err**2)/npix
#     #         self.starchisq[i] = chisq
#     #         # chi value, RMS of the residuals as a fraction of the amp
#     #         rms = np.sqrt(np.mean(((flux-model.ravel())/self.staramp[i])**2))
#     #         self.starrms[i] = rms
            
#     #         #model = psf(x,y,pars=[amp,xcen,ycen])
#     #         # Zero-out anything beyond the fitting radius
#     #         #im[mask] = 0.0
#     #         #npix = im.size
#     #         #npix = len(x)
#     #         allim[pixcnt:pixcnt+npix] = model.flatten()
#     #         pixcnt += npix

#     #         #import pdb; pdb.set_trace()
            
#     #     self.niter += 1
            
#     #    return allim


#     def fitstars(self,x,psfparams):
#         """ Fit amp/xcen/ycen for all the stars using the current PSF."""

#         # Loop over the stars and generate the model image
#         lookup = np.zeros((1,1,1),float)
#         allmodel = np.zeros(self.star_imdata.shape,float)
#         maxpix = self.star_imdata.shape[1]
#         npix = int(np.sqrt(maxpix))
#         for i in range(self.nstars):
#             im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
#             pars = np.zeros(3,float)
#             pars[0] = self.staramp[i]
#             pars[1] = self.starxcen[i]
#             pars[2] = self.starycen[i]
#             pout = mnb.psffit(im1,err1,xdata1,ydata1,pars,self.psftype,psfparams,
#                               lookup,self.imshape,self.verbose)
#             bestpar,perror,cov,flux,fluxerr,chisq = pout
#             rchisq = chisq/len(xdata1)
#             model,der = mnb.psf(xdata1,ydata1,pars,self.psftype,psfparams,
#                                 lookup,self.imshape)
#             rms = np.sqrt(np.mean(((im1-model)/bestpar[0])**2))
#             #print(i,bestpar,rchisq,rms)
            
#             # Update stellar parameters
#             self.staramp[i] = bestpar[0]
#             self.starxcen[i] = bestpar[1]
#             self.starycen[i] = bestpar[2]
#             self.starchisq[i] = rchisq
#             self.starrms[i] = rms
            
    
#     def jac(self,x,args):
#         """ jacobian."""
#         # input the model parameters        
#         if self.verbose:
#             print('model: '+str(self.niter)+' ',args)
#         # Loop over the stars and generate the model image
#         nparsarr = np.array([6,7,8,8,7])
#         npars = nparsarr[self.psftype-1]
#         allmodel = np.zeros(self.star_imdata.shape,float)
#         maxpix = self.star_imdata.shape[1]
#         allderiv = np.zeros((self.nstars,maxpix,npars),float)
#         for i in range(self.nstars):
#             im1,err1,xdata1,ydata1,bbox1,n1 = self.unpackfitstar(i)
#             allpars = np.zeros(npars,float)
#             allpars[0] = self.staramp[i]
#             allpars[1] = self.starxcen[i]
#             allpars[2] = self.starycen[i]
#             allpars[3:] = args
#             g,d = mnb.amodel2d(xdata1,ydata1,self.psftype,allpars,npars)
#             allmodel[i,:len(g)] = g
#             allderiv[i,:len(g),:] = d
#         # Trim off the stellar parameter (amp/xc/yc) derivatives
#         allderiv = allderiv[:,:,3:]
#         return allmodel,allderiv

#     def linesearch(self,bestpar,dbeta):
#         # Perform line search along search gradient

#         # Inside model() the parameters are limited to the PSF bounds()
#         f0 = self.chisq(bestpar)
#         f1 = self.chisq(bestpar+0.5*dbeta)
#         f2 = self.chisq(bestpar+dbeta)
#         alpha = quadratic_bisector(np.array([0.0,0.5,1.0]),np.array([f0,f1,f2]))
#         alpha = np.minimum(np.maximum(alpha,0.0),1.0)  # 0<alpha<1
#         if np.isfinite(alpha)==False:
#             alpha = 1.0

#         pars_new = bestpar + alpha * dbeta
#         new_dbeta = alpha * dbeta
#         return alpha,new_dbeta
    
#     # def mklookup(self,order=0):
#     #     """ Make an empirical look-up table for the residuals."""

#     #     # Make the empirical EPSF
#     #     cube = self.psf.resid(self.tab,self.image,fillvalue=np.nan)
#     #     coords = (self.tab['x'].data,self.tab['y'].data)
#     #     epsf,nbadstar,rms = mkempirical(cube,order=order,coords=coords,shape=self.image.shape,lookup=True)
#     #     lookup = models.PSFEmpirical(epsf,imshape=self.image.shape,order=order,lookup=True)

#     #     # DAOPHOT does some extra analysis to make sure the flux
#     #     # in the residual component is okay

#     #     # -make sure
#     #     #  -to take the total flux into account (not varying across image)
#     #     #  -make sure the amp=1 at center
#     #     #  -make sure all PSF values are >=0
                             
#     #     # Add the lookup table to the PSF model
#     #     self.psf.lookup = lookup

#     #     #import pdb; pdb.set_trace()
        
        
#     def starmodel(self,istar,params):
#         """ Generate a 2D star model image that can be compared to the original cutout."""
#         im,err,xdata,ydata,bbox,shape = self.unpackstar(istar)
#         xdata1d = xdata.ravel()
#         ydata1d = ydata.ravel()
#         amp = self.staramp[istar]
#         xcen = self.starxcen[istar]   
#         ycen = self.starycen[istar]
#         nparsarr = np.array([6,7,8,8,7])
#         npars = nparsarr[self.psftype-1]
#         allpars = np.zeros(npars,float)
#         allpars[0] = amp
#         allpars[1] = xcen
#         allpars[2] = ycen
#         allpars[3:] = params
#         g,_ = mnb.amodel2d(xdata1d,ydata1d,self.psftype,allpars,0)
#         g = g.reshape(im.shape)
#         return g


# cdef fitpsf(psftype,psfparams,image,error,starx,stary,starflux,fitradius,method='qr',maxiter=10,
#            minpercdiff=1.0,verbose=False):
#     """
#     Fit PSF model to stars in an image.

#     Parameters
#     ----------
#     psftype : PSF object
#        PSF object with initial parameters to use.
#     psfparams : numpy array
#        PSF object with initial parameters to use.
#     image : numpy array
#        Image to use to fit PSF model to stars.
#     error : numpy array
#        Uncertainties in image.
#     tab : numpy structured array
#        Table with initial amp/x/y values for the stars to use to fit the PSF.
#     fitradius : float, table
#        The fitting radius.  If none is input then the initial PSF FWHM will be used.
#     method : str, optional
#        Method to use for solving the non-linear least squares problem: "qr",
#        "svd", "cholesky", and "curve_fit".  Default is "qr".
#     maxiter : int, optional
#        Maximum number of iterations to allow.  Only for methods "qr", "svd", and "cholesky".
#        Default is 10.
#     minpercdiff : float, optional
#        Minimum percent change in the parameters to allow until the solution is
#        considered converged and the iteration loop is stopped.  Only for methods
#        "qr" and "svd".  Default is 1.0.
#     verbose : boolean, optional
#        Verbose output.

#     Returns
#     -------
#     pars : numpy array
#        Array of best-fit model parameters
#     perror : numpy array
#        Uncertainties in "pars".
#     pcov : numpy array
#        Covariances of the best-ft parameters.
#     psftab : table
#        Table of best-fitting amp/xcen/ycen values for the PSF stars.
#     rchisq : float
#        Reduced chi-squared of the best-fit.
#     pf : PSFFitter object
#        PSFFitter object.
    
#     Example
#     -------

#     pars,perror,pcov,psftab,rchisq,pf = fitpsf(psftype,psfparams,image,error,tab,fitradius)

#     """

#     # Fitting the PSF to the stars
#     #-----------------------------

#     # Empirical PSF - done differently
#     if psf.psftype==6:
#         cube1 = starcube(tab,image,npix=psf.npix,fillvalue=np.nan)
#     #     coords = (tab['x'].data,tab['y'].data)
#     #     epsf1,nbadstar1,rms1 = mkempirical(cube1,order=psf.order,coords=coords,shape=psf._shape)
#     #     initpsf = mnb.PSFEmpirical(epsf1,imshape=image.shape,order=psf.order)
#     #     pf = PSFFitter(initpsf,image,tab,fitradius=fitradius,verbose=False)
#     #     # Fit the amp, xcen, ycen properly
#     #     xdata = np.arange(pf.ntotpix)
#     #     out = pf.model(xdata,[])
#     #     # Put information into the psftab table
#     #     psftab['amp'] = pf.staramp
#     #     psftab['x'] = pf.starxcen
#     #     psftab['y'] = pf.starycen
#     #     psftab['chisq'] = pf.starchisq
#     #     psftab['rms'] = pf.starrms
#     #     psftab['npix'] = pf.starnpix    
#     #     for i in range(len(tab)):
#     #         bbox = pf.bboxdata[i]
#     #         psftab['ixmin'][i] = bbox.ixmin
#     #         psftab['ixmax'][i] = bbox.ixmax
#     #         psftab['iymin'][i] = bbox.iymin
#     #         psftab['iymax'][i] = bbox.iymax        
#     #     psftab = Table(psftab)
#     #     # Remake the empirical EPSF    
#     #     cube = starcube(psftab,image,npix=psf.npix,fillvalue=np.nan)
#     #     epsf,nbadstar,rms = mkempirical(cube,order=psf.order,coords=coords,shape=psf._shape)
#     #     newpsf = mnb.PSFEmpirical(epsf,imshape=image.shape,order=psf.order)
#     #     if verbose:
#     #         print('Median RMS: '+str(np.median(pf.starrms)))
#     #         print('dt = %.2f sec' % (time.time()-t0))
#     #     return newpsf, None, None, psftab, pf

#     nstars = len(starx)
#     pf = PSFFitter(psftype,psfparams,image,error,starx,stary,starflux,fitradius,verbose)
#     initpar = psfparams.copy()
#     xdata = np.arange(10)  # dummy

#     # Iterate
#     niter = 0
#     percdiff = 1e10
#     bestpar = initpar.copy()
    
#     dchisq = -1
#     oldchisq = 1e30
#     bounds = mnb.model2d_bounds(psftype)
#     bounds = bounds[3:,:]
#     while (niter<maxiter and percdiff>minpercdiff and dchisq<0):
#         # Fit the stellar parameters first
#         pf.fitstars(xdata,bestpar)
#         # Get the Jacobian and model
#         m,jac = pf.jac(xdata,bestpar)
#         chisq = np.sum((pf.starfit_imdata-m)**2/pf.starfit_errdata**2)
#         dy = (pf.starfit_imdata-m).ravel()
#         # Weights
#         wt = 1/pf.starfit_errdata**2
#         wt = wt.ravel()
#         jac = jac.copy().reshape((jac.shape[0]*jac.shape[1],jac.shape[2]))
#         m = m.ravel()
#         # Solve Jacobian
#         dbeta = mnb.qr_jac_solve(jac,dy,weight=wt)

#         # Perform line search
#         alpha,new_dbeta = pf.linesearch(bestpar,dbeta)
            
#         if verbose:
#             print('  pars = '+str(bestpar))
#             print('  dbeta = '+str(dbeta))

#         # Update parameters
#         oldpar = bestpar.copy()
#         # limit the steps to the maximum step sizes and boundaries
#         maxsteps = mnb.model2d_maxsteps(psftype,bestpar)
#         maxsteps = maxsteps[3:]
#         bestpar = mnb.newlsqpars(bestpar,new_dbeta,bounds,maxsteps)
        
#         # Update the parameters 
#         diff = np.abs(bestpar-oldpar)
#         denom = np.abs(oldpar.copy())
#         bad, = np.where(denom==0)
#         if len(bad)>0:
#             denom[bad] = 1.0  # deal with zeros
#         percdiff = np.max(diff/denom*100)
#         dchisq = chisq-oldchisq
#         percdiffchisq = dchisq/oldchisq*100
#         oldchisq = chisq
#         niter += 1
            
#         if verbose:
#             print('  '+str(niter+1)+' ',bestpar,' '+str(percdiff)+' '+str(chisq))

#     # Fit the stellar parameters first
#     pf.fitstars(xdata,bestpar)
#     # Get the Jacobian and model
#     m,jac = pf.jac(xdata,bestpar)
#     chisq = np.sum((pf.starfit_imdata-m)**2/pf.starfit_errdata**2)
#     dy = (pf.starfit_imdata-m).ravel()
#     # Weights
#     wt = 1/pf.starfit_errdata**2
#     wt = wt.ravel()
#     jac = jac.copy().reshape((jac.shape[0]*jac.shape[1],jac.shape[2]))
#     rchisq = chisq/np.sum(pf.starfit_ndata)
    
#     # Estimate uncertainties
#     # Calculate covariance matrix
#     pcov = mnb.jac_covariance(jac,dy,wt=wt)
#     perror = np.sqrt(np.diag(pcov))

#     pars = bestpar
#     if verbose:
#         print('Best-fitting parameters: '+str(pars))
#         print('Errors: '+str(perror))
#         print('Median RMS: '+str(np.median(pf.starrms)))

#     # Output best-fitting values for the PSF stars as well
#     psftab = np.zeros((nstars,11),float)
#     psftab[:,0] = np.arange(nstars)+1.0  # id
#     psftab[:,1] = pf.staramp             # amp
#     psftab[:,2] = pf.starxcen            # x
#     psftab[:,3] = pf.starycen            # y
#     psftab[:,4] = pf.starfit_ndata       # npix
#     psftab[:,5] = pf.starrms             # rms
#     psftab[:,6] = pf.starchisq           # chisq
#     psftab[:,7] = pf.star_bbox[:,0]      # ixmin
#     psftab[:,8] = pf.star_bbox[:,1]      # ixmax
#     psftab[:,9] = pf.star_bbox[:,2]      # iymin
#     psftab[:,10] = pf.star_bbox[:,3]     # iymax
    
#     return pars, perror, pcov, psftab, rchisq
    

# cdef getpsf(psftype,psfparams,image,error,starx,stary,starflux,fitradius,
#            lookup=False,lorder=0,subnei=False,
#            alltab=None,maxiter=10,minpercdiff=1.0,reject=False,maxrejiter=3,verbose=False):
#     """
#     Fit PSF model to stars in an image with outlier rejection of badly-fit stars.

#     Parameters
#     ----------
#     psftype : PSF object
#        PSF object with initial parameters to use.
#     psfparams : numpy array
#        PSF object with initial parameters to use.
#     image : numpy array
#        Image to use to fit PSF model to stars.
#     error : numpy array
#        Uncertainties in image.
#     tab : table
#        Catalog with initial amp/x/y values for the stars to use to fit the PSF.
#     fitradius : float, table
#        The fitting radius.  If none is input then the initial PSF FWHM will be used.
#     lookup : boolean, optional
#        Use an empirical lookup table.  Default is False.
#     lorder : int, optional
#        The order of the spatial variations (0=constant, 1=linear).  Default is 0.
#     method : str, optional
#        Method to use for solving the non-linear least squares problem: "qr",
#        "svd", "cholesky", and "curve_fit".  Default is "qr".
#     subnei : boolean, optional
#        Subtract stars neighboring the PSF stars.  Default is False.
#     alltab : table, optional
#        Catalog of all objects in the image.  This is needed for bad PSF star
#        rejection.
#     maxiter : int, optional
#        Maximum number of iterations to allow.  Only for methods "qr", "svd", and "cholesky".
#        Default is 10.
#     minpercdiff : float, optional
#        Minimum percent change in the parameters to allow until the solution is
#        considered converged and the iteration loop is stopped.  Only for methods
#        "qr" and "svd".  Default is 1.0.
#     reject : boolean, optional
#        Reject PSF stars with high RMS values.  Default is False.
#     maxrejiter : int, boolean
#        Maximum number of PSF star rejection iterations.  Default is 3.
#     verbose : boolean, optional
#        Verbose output.

#     Returns
#     -------
#     newpsf : PSF object
#        New PSF object with the best-fit model parameters.
#     pars : numpy array
#        Array of best-fit model parameters
#     perror : numpy array
#        Uncertainties in "pars".
#     psftab : table
#        Table of best-fitting amp/xcen/ycen values for the PSF stars.

#     Example
#     -------

#     newpsf,pars,perror,psftab = getpsf(psf,image,tab)

#     """

#     #t0 = time.time()
#     #print = utils.getprintfunc() # Get print function to be used locally, allows for easy logging   

#     #psftype,psfpars,_,_ = mnb.unpackpsf(psf)
    
#     # # Fitting radius
#     # if fitradius is None:
#     #     tpars = np.zeros(len(psfpars)+3,float)
#     #     tpars[0] = 1.0
#     #     tpars[3:] = psfpars
#     #     if psftype == 3:  # Penny
#     #         fitradius = mnb.penny2d_fwhm(tpars)*1.5
#     #     else:
#     #         fitradius = mnb.model2d_fwhm(psftype,tpars)
        
#     # # subnei but no alltab input
#     # if subnei and alltab is None:
#     #     raise ValueError('alltab is needed for PSF neighbor star subtraction')
        
#     # if 'id' not in tab.dtype.names:
#     #     tab['id'] = np.arange(len(tab))+1
#     # psftab = tab.copy()

#     # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
#     psftab = np.zeros((nstars,12),float)
#     psftab[:,0] = np.arange(nstars)+1
#     psftab[:,1] = starflux
#     psftab[:,2] = starx
#     psftab[:,3] = stary
    
#     # Initializing output PSF star catalog
#     #dt = np.dtype([('id',int),('amp',float),('x',float),('y',float),('npix',int),
#     #               ('rms',float),('chisq',float),('ixmin',int),('ixmax',int),
#     #               ('iymin',int),('iymax',int),('reject',int)])
#     #outtab = np.zeros(len(tab),dtype=dt)
#     #outtab = Table(outtab)
#     #for n in ['id','x','y']:
#     #    outtab[n] = tab[n]
#     nstars = len(starx)
#     # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
#     outtab = np.zeros((nstars,12),float)
    
#     # Remove stars that are too close to the edge
#     ny,nx = image.shape
#     gd, = np.where((starx>fitradius) & (starx<(nx-1-fitradius)) &
#                    (stary>fitradius) & (stary<(ny-1-fitradius)))
#     ngd = len(gd)
#     nbd = nstars-ngd
#     if nbd > 0:
#         if verbose:
#             print('Removing '+str(nbd)+' stars near the edge')
#         psftab = psftab[gd,:]

#     # # Generate an empirical image of the stars
#     # # and fit a model to it to get initial estimates
#     # if psftype != 6:
#     #     cube = starcube(psftab,image,npix=psf.npix,fillvalue=np.nan)
#     #     epsf,nbadstar,rms = mkempirical(cube,order=0)
#     #     #epsfim = CCDData(epsf,error=epsf.copy()*0+1,mask=~np.isfinite(epsf))
#     #     epsfim = epsf.copy()
#     #     epsferr = np.ones(epsf.shape,float)
#     #     ny,nx = epsf.shape
#     #     xx,yy = np.meshgrid(np.arange(nx),np.arange(ny))
#     #     out = model2dfit(epsfim,epsferr,xx,yy,psftype,1.0,nx//2,ny//2,verbose=False)
#     #     pars,perror,cov,flux,fluxerr,chisq = out
#     #     mparams = pars[3:]  # model parameters
#     #     #pars,perror,mparams = mnb.model2d_fit(epsfim,pars=[1.0,psf.npix/2,psf.npix//2])
#     #     initpar = mparams.copy()
#     #     curpsf = mnb.packpsf(psftype,mparams,0,0)
#     #     #curpsf = psf.copy()
#     #     #curpsf.params = initpar
#     #     if verbose:
#     #         print('Initial estimate from empirical PSF fit = '+str(mparams))
#     # else:
#     #     curpsf = psf.copy()
#     #     _,initpar,_,_ = mnb.unpackpsf(psf)
#     #     #initpar = psf.params.copy()
#     curparams = psfparams.copy()
    
#     # Outlier rejection iterations
#     nrejiter = 0
#     flag = 0
#     nrejstar = 100
#     fitrad = fitradius
#     useimage = image.copy()
#     while (flag==0):
#         if verbose:
#             print('--- Iteration '+str(nrejiter+1)+' ---')                

#         # Update the fitting radius
#         #if nrejiter>0:
#         #    fitrad = curpsf.fwhm()
#         if verbose:
#             print('  Fitting radius = %5.3f' % (fitrad))
                    
#         # Reject outliers
#         if reject and nrejiter>0:
#             medrms = np.median(ptab[:,5]) # rms
#             sigrms = utils.mad(ptab[:,5])
#             gd, = np.where(ptab[:,5] < medrms+3*sigrms)
#             nrejstar = len(psftab)-len(gd)
#             if verbose:
#                 print('  RMS = %6.4f +/- %6.4f' % (medrms,sigrms))
#                 print('  Threshold RMS = '+str(medrms+3*sigrms))
#                 print('  Rejecting '+str(nrejstar)+' stars')
#             if nrejstar>0:
#                 psftab = psftab[gd,:]

#         # Subtract neighbors
#         if nrejiter>0 and subnei:
#             if verbose:
#                 print('Subtracting neighbors')
#                 # Find the neighbors in alltab
#                 # Fit the neighbors and PSF stars
#                 # Subtract neighbors from the image
#                 useimage = image.copy()  # start with original image
#                 print('subtractnei')
#                 #useimage = subtractnei(useimage,alltab,tab,curpsf)
                
#         # Fitting the PSF to the stars
#         #-----------------------------
#         newpars,perror,pcov,ptab,rchisq = fitpsf(psftype,curparams,image,error,psftab[:,2],psftab[:,3],
#                                                  psftab[:,1],fitradius,method=method,
#                                                  maxiter=maxiter,minpercdiff=minpercdiff,verbose=verbose)
#         #newpsf,pars,perror,ptab,pf = fitpsf(curpsf,useimage,psftab,fitradius=fitrad,method=method,
#         #                                    maxiter=maxiter,minpercdiff=minpercdiff,verbose=verbose)
        
#         # Add information into the output catalog
#         _,ind1,ind2 = np.intersect1d(outtab[:,0],ptab[:,0],return_indices=True)
#         #ind1,ind2 = dln.match(outtab['id'],ptab['id'])
#         outtab[:,11] = 1    # reject
#         #outtab['reject'] = 1
#         #for n in ptab.columns:
#         #    outtab[n][ind1] = ptab[n][ind2]
#         #outtab['reject'][ind1] = 0
#         # id, amp, x, y, npix, rms, chisq, ixmin, ixmax, iymin, iymax, reject
#         outtab[ind1,1:10] = ptab[ind2,1:10]
        
#         # Compare PSF parameters
#         # if type(newpsf)!=mnb.PSFEmpirical:
#         #     pardiff = newpsf.params-curpsf.params
#         # else:
#         #     pardiff = newpsf._data-curpsf._data
#         pardiff = newpars - pars
#         sumpardiff = np.sum(np.abs(pardiff))
#         curpsf = newpsf.copy()
        
#         # Stopping criteria
#         if reject is False or sumpardiff<0.05 or nrejiter>=maxrejiter or nrejstar==0: flag=1
#         if subnei is True and nrejiter==0: flag=0   # iterate at least once with neighbor subtraction
        
#         nrejiter += 1
        
#     # # Generate an empirical look-up table of corrections
#     # if lookup:
#     #     if verbose:
#     #         print('Making empirical lookup table with order='+str(lorder))

#     #     pf.mklookup(lorder)
#     #     # Fit the stars again and get new RMS values
#     #     xdata = np.arange(pf.ntotpix)
#     #     out = pf.model(xdata,*pf.psf.params)
#     #     newpsf = pf.psf.copy()
#     #     # Update information in the output catalog
#     #     ind1,ind2 = dln.match(outtab['id'],ptab['id'])
#     #     outtab['reject'] = 1
#     #     outtab['reject'][ind1] = 0
#     #     outtab['amp'][ind1] = pf.staramp[ind2]
#     #     outtab['x'][ind1] = pf.starxcen[ind2]
#     #     outtab['y'][ind1] = pf.starycen[ind2]
#     #     outtab['rms'][ind1] = pf.starrms[ind2]
#     #     outtab['chisq'][ind1] = pf.starchisq[ind2]                
#     #     if verbose:
#     #         print('Median RMS: '+str(np.median(pf.starrms)))    
            
    
#     return newpsf, pars, perror, outtab
        
        
